# ***************************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************

from cli.commands.cli_command_provider import CliCommandProvider
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NODE_TYPE_CONTROLLER
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import get_node_type
from cli.utils.signal_utils import send_sighup

import errno
import json
from mako import compat
import nsx_logging
import os
import cli.commands.host_shared.ovs_util as ovs_util

nsx_exporter_map = {
    NODE_TYPE_EDGE: "nsx-edge-exporter",
    NODE_TYPE_AUTONOMOUS_EDGE: "nsx-edge-exporter",
    NODE_TYPE_HOST_SLESKVM: "nsx-exporter",
    NODE_TYPE_HOST_CENTOSKVM: "nsx-exporter",
    NODE_TYPE_HOST_RHELKVM: "nsx-exporter",
    NODE_TYPE_HOST_UBUNTUKVM: "nsx-exporter",
    NODE_TYPE_MANAGER: "host-node-status-reporter",
    NODE_TYPE_CONTROLLER: "host-node-status-reporter",
    NODE_TYPE_HOST_ESX: "nsx-exporter",
    NODE_TYPE_BAREMETAL_UBUNTU_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_RHEL_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_CENTOS_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_SLES_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_OEL_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_WINDOWS_SERVER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_RHEL_CONTAINER: "nsx-exporter",
    NODE_TYPE_BAREMETAL_CENTOS_CONTAINER: "nsx-exporter",
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY: "nsx-edge-exporter"
}

lg = nsx_logging.getLogger("cli.commands.common.exporter_log_level")

# Use nsx-manager if exists, else use the first node type
node_types = get_node_type().split()
if NODE_TYPE_MANAGER in node_types:
    g_node_type = NODE_TYPE_MANAGER
else:
    g_node_type = node_types[0]

LOGGING_LEVEL_MAP = {
    "off": "0",
    "fatal": "1",
    "error": "2",
    "warn": "3",
    "info": "4",
    "debug": "5",
    "trace": "6"
}
# when logging-level set through dfw-exporter uds server, we need transfer the str to
# dfw-exporter logging-level friendly noun.
log_level_str_map = {
    "off": "none",
    "fatal": "panic",
    "warn": "warning",
    "debug": "verbose",
    "trace": "trivia"
}

# ----------------------------------------------------------------------
# set exporter service logging level
# ----------------------------------------------------------------------
class SetExporterLoggingLevel(CliCommandProvider):
    def __call__(self, cmd_list):
        service_name = nsx_exporter_map[g_node_type]
        level = self.get_arg_value("logging-level-arg", cmd_list)
        if self.set_logging_level(service_name, level) == True:
            return "OK\n\n%s" % json.dumps("")
        else:
            return "MSG_SET_LOG_LEVEL_FAILED"

    def update_config_file(self, daemon_name, level):
        #  Updates daemon's config file with specified log level
        filename = "/etc/vmware/"+ daemon_name + "/config.json"
        logLevel = LOGGING_LEVEL_MAP[level]
        if os.path.exists(os.path.dirname(filename)):
            with open(filename, 'r+') as f:
                data = json.load(f)
                data['logLevel'] = logLevel
                f.seek(0)
                json.dump(data, f, indent=4)
        else:
            if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
                               NODE_TYPE_CONTROLLER, NODE_TYPE_MANAGER,
                               NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                lg.error("Config file %s not found", filename, error_code=ec.OPERATION_FAILED)
                return False

            try:
                original_umask = os.umask(0)
                os.makedirs(os.path.dirname(filename), compat.octal("0775"))
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    lg.error("Could not create config dir %s exception %s",
                                os.path.dirname(filename), exc, error_code=ec.OPERATION_FAILED)
                    return False
            finally:
                os.umask(original_umask)
            with open(filename, 'w') as f:
                data = {}
                data["logLevel"] = logLevel
                json.dump(data, f, indent=4)

        return True

    def set_logging_level(self, daemon_name, level):
        if g_node_type in [NODE_TYPE_BAREMETAL_WINDOWS_SERVER]:
        #  For windows signal stragey is not enable at agg-service, so we
        #  leverage dfw uds cliServer to set nsx-exporter logging-level.
            dfw_level = log_level_str_map.get(level, level)
            reponse = ovs_util.run_exporter_cli("set/loglevel", dfw_level)
            if reponse is None:
                lg.error("Failed to call dfw_exporter vertical cliServer %s",
                         daemon_name, error_code=ec.OPERATION_FAILED)
                return False
            else:
                return True
        else:
        #  Updates daemon's config file with specified log level. Then sends
        #  SIGHUP signal to the daemon.
        #  Returns true if successful, otherswise returns false.
            if not self.update_config_file(daemon_name, level):
                return False

            if send_sighup(daemon_name):
                return True
            else:
                lg.error("Failed to send SIGHUP to service %s", daemon_name, error_code=ec.OPERATION_FAILED)
                return False

