# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************


__all__ = ["CliCommandDescriptor"]


import codecs
import getpass
import nsx_logging
import os
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import subprocess
import sys
import time
import yaml

from lib.json_schema import JsonSchema
from cli.descriptors.cli_argument_descriptor import is_argument
from cli.descriptors.cli_argument_descriptor import resolve_argument_name
from cli.descriptors.descriptor import Descriptor
from cli.descriptors.descriptor import SCHEMA_REGISTRY
from cli.descriptors.error import DescriptorException
from cli.utils.central_cli_api import _invoke_and_wait
from cli.utils.central_cli_utils import is_central_cli_command
from cli.utils.central_cli_utils import get_central_cli_nodes
from cli.utils.central_cli_utils import get_central_cli_user_input
from cli.utils.central_cli_utils import get_cmd_list_path_tuple
from cli.utils.constants import CENTRAL_CLI_STR
from cli.utils.constants import CMD_MATCH_EXACT
from cli.utils.constants import CMD_MATCH_FAILED
from cli.utils.constants import CMD_MATCH_INCOMPLETE
from cli.utils.constants import CMD_MATCH_PARTIAL
from cli.utils.constants import CMD_MATCH_VALUE_MISMATCH
from cli.utils.constants import OUTPUT_MODIFIER_TOKEN
from cli.utils.constants import OUTPUT_MODIFIER_COUNT
from cli.utils.constants import OUTPUT_MODIFIER_FIND
from cli.utils.constants import OUTPUT_MODIFIER_FIRST
from cli.utils.constants import OUTPUT_MODIFIER_IGNORE
from cli.utils.constants import OUTPUT_MODIFIER_JSON
from cli.utils.constants import OUTPUT_MODIFIER_LAST
from cli.utils.constants import OUTPUT_MODIFIER_MORE
from cli.utils.constants import OUTPUT_MODIFIER_SORT
from cli.utils.constants import PERMISSION_CRUD
from cli.utils.constants import PERMISSION_EXECUTE
from cli.utils.constants import PERMISSION_READ
from cli.utils.constants import PERMISSION_NONE
from cli.utils.constants import PROVIDER_TARGET_ANY
from cli.utils.error_utils import render_error
from cli.utils.error_utils import render_error_result
from cli.utils.import_utils import get_class
from cli.utils.node_utils import get_node_type
from cli.utils.render_utils import render_mako_template_str,\
                                generate_datetime_stamp_str
from cli.utils.error_code import ErrorCode as ec


lg = nsx_logging.getLogger("cli.descriptors.cli_command_descriptor")


CLI_BASE_DIR = os.path.dirname(os.path.dirname(__file__))
CLI_COMMAND_SCHEMA_FILE = os.path.join(os.path.dirname(__file__),
                                       'cli_command_descriptor.yml')
CLI_COMMAND_TYPES = dict((r.id, r)
                         for r in yaml.load(open(CLI_COMMAND_SCHEMA_FILE), Loader=yaml.Loader))
for key in list(CLI_COMMAND_TYPES.keys()):
    if key not in SCHEMA_REGISTRY:
        JsonSchema(CLI_COMMAND_TYPES[key].raw_schema,
                   schema_registry=SCHEMA_REGISTRY, type_path=[key],
                   strict=True)


CLI_REQUIRED_PERMISSIONS = [
    PERMISSION_CRUD, PERMISSION_EXECUTE, PERMISSION_READ, PERMISSION_NONE]
# User with permission, key, can invoke commands with permissions in values
CLI_AUTHORIZED_PERMISSIONS = {
    PERMISSION_CRUD: CLI_REQUIRED_PERMISSIONS,
    PERMISSION_EXECUTE: [PERMISSION_EXECUTE, PERMISSION_READ, PERMISSION_NONE],
    PERMISSION_READ: [PERMISSION_READ, PERMISSION_NONE],
    PERMISSION_NONE: [PERMISSION_NONE]
}


class CliCommandDescriptor(Descriptor):
    """Descriptor specification of a CLI Command."""
    yaml_tag = '!CliCommand'

    _provider_class = {
        "mock": "MockProvider",
        "inline_shell_script": "InlineShellScriptProvider",
        "local_shell_script": "LocalShellScriptProvider",
        "local_python": "LocalPythonProvider"}

    _renderer_class = {
        "inline_mako": "InlineMakoRenderer",
        "local_mako_file": "LocalMakoFileRenderer"}

    def __init__(self, id, schema_spec):
        super(CliCommandDescriptor, self).__init__(id, schema_spec)
        self._node_type = get_node_type()
        self._service = None  # Set by factory
        self._raw_schema = schema_spec
        self._provider = self._resolve_provider()
        self._renderer = self._resolve_renderer()
        self._path_tuple = tuple(self._raw_schema["path"])
        self._hidden = self._raw_schema.get("hidden", False)
        self._doc_hidden = self._raw_schema.get("doc_hidden", False)
        self._doc_override = self._raw_schema.get("doc_override", False)
        self._debug = self._raw_schema.get("debug", False)
        self._context_available = self._raw_schema.get(
            "context_available", ["all"])
        self._context_command = self._raw_schema.get("context_command", False)
        self._show_configuration_category = self._raw_schema.get(
            "show_configuration_category")
        self._suppress_modifiers = self._raw_schema.get(
            "suppress_modifiers", False)
        self._additional_last_args = self._raw_schema.get(
            "additional_last_args", 0)
        interactive_args = [InteractiveArgument(**arg)
                            for arg in schema_spec.get("interactive_args", [])]
        self._interactive_args = dict((arg.name, arg)
                                      for arg in interactive_args)
        self._nsx_feature = self._raw_schema.get("nsx_feature")
        if "required_permissions" not in self._raw_schema:
            raise DescriptorException(
                "CliCommand descriptor '%s' missing required_permissions "
                "property" % id)
        self._required_permissions = self._raw_schema["required_permissions"]
        if self._required_permissions not in CLI_REQUIRED_PERMISSIONS:
            raise DescriptorException(
                "Invalid required_permissions: %s" % self._required_permissions)
        self._authorized = False  # Set by DescriptorFactory during yml load
        self._requires_tty_if_remote = self._raw_schema.get("requires_tty_if_remote", False)

        # What node types, this descriptor is applicable for. This gets set
        # within descriptor_factory during module load phase. e.g.
        # self.node_types = ['EdgeNode', 'ManagementClusterRoleConfig']
        # In above case, the command is applicable to run on Edge node and
        # on manager, when invoked using Central CLI
        self.node_types = []

    @property
    def raw_schema(self):
        return self._raw_schema

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, value):
        self._service = value

    @property
    def hidden(self):
        return self._hidden

    @property
    def doc_hidden(self):
        return self._doc_hidden

    @property
    def doc_override(self):
        return self._doc_override

    @property
    def debug(self):
        return self._debug

    @property
    def interactive_args(self):
        return self._interactive_args

    @property
    def context_available(self):
        return self._context_available

    @property
    def context_command(self):
        return self._context_command

    @property
    def show_configuration_category(self):
        return self._show_configuration_category

    @property
    def provider(self):
        return self._provider

    @property
    def suppress_modifiers(self):
        return (self._suppress_modifiers or
                self._provider.watch_provider if self._provider else False)

    @property
    def additional_last_args(self):
        return self._additional_last_args

    @property
    def watch_provider(self):
        return self._provider.watch_provider if self._provider else False

    @property
    def watch_process(self):
        return self._provider.watch_process if self._provider else False

    @property
    def multi_args(self):
        factory = self.service._factory
        result = []
        for a in self.path_tuple:
            desc = factory.arguments.get(resolve_argument_name(a))
            if desc:
                if desc.limit > 1:
                    result.append(a)
            else:
                continue

        return result

    def get_arg_indexes(self, cmd_list):
        try:
            _factory = self.service._factory
        except AttributeError:
            _factory = None
        arg_indexes = {}
        offset = 0
        for index, key in enumerate(self.path_tuple):
            if is_argument(key):
                if not _factory:
                    arg_ending_word = None
                else:
                    arg_desc = _factory.arguments.get(resolve_argument_name(key))
                    # arg_desc = _factory.arguments.get(key)
                    arg_limit = arg_desc.limit
                    arg_ending_word = arg_desc.ending_word

                if not arg_ending_word:
                    indexes = arg_indexes.setdefault(resolve_argument_name(key), [])
                    indexes.append(offset)
                else:
                    while (cmd_list[offset] != arg_ending_word):
                        indexes = arg_indexes.setdefault(resolve_argument_name(key), [])
                        indexes.append(offset)
                        offset += 1
            offset += 1

        return arg_indexes


    @property
    def required_permissions(self):
        return self._required_permissions

    @property
    def authorized(self):
        return self._authorized

    @property
    def requires_tty_if_remote(self):
        return self._requires_tty_if_remote

    @property
    def path_tuple(self):
        # If running in session mode, remove the first 2 items (i.e. 'on' and
        # '<registered-node-uuid>' from descriptor's _path_tuple
        try:
            if (self.service.session_mode_active and
                self.id.endswith(CENTRAL_CLI_STR) and
                self.id != CENTRAL_CLI_STR):
                return self._path_tuple[2:]
            else:
                return self._path_tuple
        except AttributeError: #  self._service may not be set for unit tests
            return self._path_tuple

    def check_authorized(self, user_permissions):
        """Set self._authorized True if current user can invoke command.

        Args:
            user_permissions: Current user's assigned permissions.
        """
        self._authorized = (
            self._required_permissions in
            CLI_AUTHORIZED_PERMISSIONS.get(user_permissions, []))

    def normalize_command_list(self, cmd_list, verify_argument=True):
        """Normalize a user entered command.

        Args:
            cmd_list: Command entered by user.
        Return:
            Return normalized command where any command shortcuts have been
            replaced by the actual command token defined by the CliCommand's
            path property.  For example: "sho pro mo" would be normalized to
            "show process monitor".
        Note:
            Before calling this method, the caller should ensure the user
            command is valid for this command descriptor by calling
            match_command_list.
        """
        # The Help command requires special handling,
        # only normalize the "help" keyword
        if self.id == "Help":
            return [self.path_tuple[0]] + cmd_list[1:]
        last_arg_name = None
        if self.additional_last_args and is_argument(self.path_tuple[-1]):
            last_arg_name = self.path_tuple[-1]
        result = []
        pos = 0
        arg_indexes = self.get_command_indexes(cmd_list)
        for idx, arg in enumerate(self.path_tuple):
            size = 1
            if is_argument(arg):
                name = resolve_argument_name(arg)
                size = arg_indexes[idx][1]
                # In case of argument value is made up of multiple words (e.g.
                # RegisteredNodeUuid used for Central CLI commands, which is a
                # multi-arg), this would construct final arg value by joining
                # all the word tokens provided for the value of given multi-arg
                arg = ' '.join(cmd_list[pos:(pos+size)])  # case sensitive
                matched_value = self.service.match_argument_value(
                                    name, arg, verify_argument)
                if matched_value is not None:
                    arg = matched_value
                # TODO: If not matched_value, then appending arg would be wrong
            result.append(arg)
            pos += size
        # If descriptor supports additional last args, add to result
        if self.additional_last_args and last_arg_name:
            name = resolve_argument_name(last_arg_name)
            while pos < len(cmd_list):
                arg = cmd_list[pos]
                matched_value = self.service.match_argument_value(
                                    name, arg, verify_argument)
                if matched_value is not None:
                    result.append(matched_value)
                pos += 1
        return result

    def invoke(self, cmd_list, local_client=True, watch_stdout=None):
        """Invoke command.

        Args:
            cmd_list: Command entered by user.
            context: Current context.
            local_client: True, if CLI is running on local server.
            watch_stdout: stdout file object to use instead of sys.stdout.
        Return:
            Output from command provider implementation.
        """
        if self._provider is None:
            sys.stdout.write("\n")
            return None
        try:
            context, context_data = self.service.get_context()
            if context_data is None:
                context_data = ""
            else:
                try:
                    context_data = simplejson.dumps(context_data)
                except:
                    context_data = str(context_data)
            modifier_str = self.service.get_modifier()
            modifier_str = "" if modifier_str is None else modifier_str
            if self.id.endswith(CENTRAL_CLI_STR):
                # TODO: consider moving this block to cli_command_service
                # or central_cli_utils as it's only executed when the user
                # invokes a context command within session mode
                if not self.service.session_mode_active:
                    nodes = get_central_cli_nodes(cmd_list)
                    cmd = ' '.join(get_central_cli_user_input(cmd_list))
                else: # In session mode
                    nodes = self.service.session_nodes
                    cmd = ' '.join(cmd_list)
                factory = self.service._factory
                nodes = [factory.node_cache[node]['uuid'] for node in nodes]

                # To be returned in case of failure/default scenario
                current_context_result = (
                    'OK\n\n"%s"\n\n"%s"' % (context, context_data))
                # Target node in Central CLI needs to know that this
                # API request is for context command. As such
                # don't produce any output, an environment variable
                # CENTRAL_CLI_STR is being appended. This will
                # signal the receiving node to always display the
                # output of context command.
                cmd = '%s=1\\t%s' % (CENTRAL_CLI_STR, cmd)
                for node in nodes:
                    try:
                        result = _invoke_and_wait(node, cmd, modifier_str,
                                    context,
                                    getpass.getuser(),
                                    [factory._user_permissions],
                                    self.debug)
                    except CentralCLIAPIException:
                        render_error("MSG_CLI_UNEXPECTED_ERROR_PROTON")
                        return current_context_result
                    else:
                        return result[2] # Next context, context_data
                return current_context_result
            else:
                return self._provider.invoke(
                    cmd_list, context, context_data, modifier_str, local_client,
                    watch_stdout)
        except KeyboardInterrupt:
            if self._provider.watch_provider:
                if watch_stdout:
                    watch_stdout.write("\n")
                else:
                    sys.stdout.write("\n")
            return None

    def render(self, output, modifier_str):
        """Render command output to stdout.

        Args:
            output: Output from command provider implementation.
            modifier_str: Output modifier entered by user.
        """
        sort_key = self._provider.get_sort_func()\
                   if hasattr(self._provider, "get_sort_func") else None
        output_prefix = None
        # Verify if status flag is enable to display datetime stamp
        # All cmds with 'get' should output datetime stamp
        # Any cmd with 'json' modifier should be ignore to print datetime stamp in output
        if self.path_tuple[0] == "get" and\
           self.service.cli_config.get('cli_output_datetime', False) and\
           not (modifier_str and modifier_str.startswith(OUTPUT_MODIFIER_JSON)):
            # If node properties have changed, then
            # get the system datetime in case if restart of nsx-cli session is required.
            start_time = None
            if not self.service.get_requires_restart():
                self.service.get_command_start_time()
            output_prefix = generate_datetime_stamp_str(start_time)
        self._renderer.invoke(output, modifier_str, sort_key, output_prefix)

    def show_configuration(self):
        """Invoke provider's show_configuration method."""
        if self._provider:
            return self._provider.show_configuration()

    def allowed_on_nodes(self, nodes):
        """Return True if one of specified nodes can run this command.

        Args:
            nodes: List of node UUIDs to check.
        Return:
            True, if one of specified nodes can run this command; else False.
        """
        # Get node_types of specified nodes from the node_cache
        nodes = [
            uuid.lower() for uuid in nodes
            if uuid.lower() in self._factory.node_cache]
        if not nodes:
            return False
        node_types = [
            v['node_type'] for k, v in list(self._factory.node_cache.items())
            if k in nodes]
        # self.node_types is configured in DescriptorFactory and is set to
        # "KVM" for descriptors loaded from host_kvm.yml. However, the
        # node_cache "node_type" value is the "resource_type" value from GET
        # /cluster|fabric/nodes and can be CENTOSKVM, RHELKVM or UBUNTUKVM. So
        # we have to map these to KVM.
        # TODO: consolidate node properties in node_utils that performs
        # these mappings currently done in descriptor_factory, cache_utils.
        KVM_RE = ('SLESKVM|CENTOSKVM|RHELKVM|UBUNTUKVM|'
                  'UBUNTUSERVER|RHELSERVER|CENTOSSERVER|SLESSERVER|'
                  'CENTOSCONTAINER|RHELCONTAINER|WINDOWSSERVER|OELSERVER')
        node_types = [re.sub(KVM_RE, 'KVM', node_type)
                      for node_type in node_types]
        # If node type of nodes on Central CLI command doesn't match
        # with node types for descriptor, return CMD_MATCH_FAILED
        if set(node_types).isdisjoint(self.node_types):
            return False
        return True

    def match_command_list(self, cmd_list_tuple, verify_argument=True,
                           input_has_trailing_space=False):
        """Determine if a user command can be processed by this command.

        Args:
            cmd_list_tuple: Command entered by user in tuple form.
        Return:
            One of the following CMD_MATCH_ values:
            - CMD_MATCH_EXACT: command matches exactly.
            - CMD_MATCH_FAILED: command does not match.
            - CMD_MATCH_INCOMPLETE: command contained too few keywords,
                but what keywords were specified matched exactly or
                partially.
            - CMD_MATCH_PARTIAL: command matches but one or more arguments
                did not match exactly (i.e. a shortcut was specified), e.g.
                "show interface eth" when eth0 is a valid interface
            - CMD_MATCH_VALUE_MISMATCH: command failed to match one or more
                arguments, e.g. "show interface foo" when foo is not valid
        """
        try:
            # If not additional_last_args but cmd_list_tuple has extra items,
            # exception will be raised from get_command_indexes
            command_indexes = self.get_command_indexes(cmd_list_tuple)
            total_items = len(command_indexes)
        except DescriptorException as e:
            return CMD_MATCH_FAILED

        # Only process descriptor if matching node_type for given cmd_list_tuple
        if self.service.session_mode_active:
            if not self.id.endswith(CENTRAL_CLI_STR) or self.id == CENTRAL_CLI_STR:
                return CMD_MATCH_FAILED
            # If node type of nodes on Central CLI command doesn't match
            # with node types for descriptor, return CMD_MATCH_FAILED
            if not self.allowed_on_nodes(self.service.session_nodes):
                return CMD_MATCH_FAILED
        elif (is_central_cli_command(cmd_list_tuple) and
            (self.id != CENTRAL_CLI_STR)):
            # node_types gets added as list to CliCommandDescriptor during
            # descriptor_factory module load time.
            nodes = get_central_cli_nodes(cmd_list_tuple)
            if not self.allowed_on_nodes(nodes):
                return CMD_MATCH_FAILED

        exact_match = True
        is_last_index_partial = False
        value_mismatch = False

        pos = 0
        next_pos = 0
        is_last_index = lambda idx: idx == len(command_indexes) - 1

        for idx, tup in enumerate(command_indexes):
            pos = next_pos
            size = tup[1]
            next_pos = pos + size
            arg_name = self.path_tuple[idx]
            if is_argument(arg_name):
                name = resolve_argument_name(arg_name)
                # user_arg should be all command words, with spaces between them
                user_arg = ' '.join(cmd_list_tuple[pos:pos+size])
                matched_value = self.service.match_argument_value(
                    name, user_arg, allow_multiple=True,
                    verify_argument=verify_argument)
                if matched_value is None:
                    if user_arg == "":
                        return CMD_MATCH_INCOMPLETE
                    exact_match = False
                    value_mismatch = True
                # Number of tokens for user_arg (applicable for multi_args only), should
                # match with matched_value. This would avoid matches like following as
                # partial matches, as they technically are incomplete matches.
                # id: RegisteredNodeUuid
                # example: user_arg = 'uuid-1 uuid-2', should not partial match with
                # 'uuid-1 uuid-2 exec'
                # TODO: Incorporate limit property of multi_args to make decision below
                elif ((user_arg != matched_value)
                        and (len(user_arg.split()) != len(matched_value.split()))):
                    return CMD_MATCH_INCOMPLETE
                # If value is `only` partially specified, not exactly matched.
                elif matched_value.startswith(user_arg) and not (matched_value == user_arg):
                    exact_match = False
                elif user_arg != matched_value:
                    # TODO: the following implementation assumes that an
                    # argument can be shortcutted, i.e. "sh interface eth"
                    # is same as "show network interface eth0" (and that there
                    # is only one interface that starts with "eth").  This
                    # will likely cause issues.
                    exact_match = False

                    # Values do not match
                    value_mismatch = True
            else:
                cmd_arg = self.path_tuple[idx]
                user_arg = cmd_list_tuple[pos]
                if cmd_arg == user_arg:
                    continue
                elif cmd_arg.startswith(user_arg):
                    exact_match = False
                    if is_last_index(idx):
                        is_last_index_partial = True
                    continue
                return CMD_MATCH_FAILED

        if value_mismatch:
            return CMD_MATCH_VALUE_MISMATCH

        # For exact match, number of tokens from user_arg
        # should be equal to required path_tuple.
        if len(command_indexes) == len(self.path_tuple):
            # If last keyword is partially matching and then there's [SPACE]
            # in user input then return no match of such command.
            # ex: > get logical-switch[HELP]
            #   => shows all commands with substring 'logical-switch'
            #   => [logical-switch, logical-switches, logical-switch-port]
            #     > get logical-switch[SPACE][HELP]
            #   => shows only command exactly matching with keyword 'logical-switch'
            if exact_match:
                return CMD_MATCH_EXACT
            else:
                if is_last_index_partial and input_has_trailing_space:
                    return CMD_MATCH_FAILED
                return CMD_MATCH_PARTIAL
        elif len(command_indexes) < len(self.path_tuple):
            if not exact_match and is_last_index_partial and\
               input_has_trailing_space:
                return CMD_MATCH_FAILED
            return CMD_MATCH_INCOMPLETE

        return CMD_MATCH_PARTIAL

    def invalid_argument_values(self, cmd_list_tuple):
        """Return list of invalid argument values specified in user command.

        Args:
            cmd_list_tuple: Command entered by user in tuple form.
        Return:
            List of invalid argument (arg_name, user_specified_arg_value)
            tuples.
        """
        result = []
        cmd_list = get_cmd_list_path_tuple(list(cmd_list_tuple))
        cmd_list_len = len(cmd_list)
        for pos in range(len(self.path_tuple)):
            arg_name = self.path_tuple[pos]
            if not is_argument(arg_name):
                continue
            elif cmd_list_len < (pos + 1):
                result.append((arg_name, None))
                continue
            name = resolve_argument_name(arg_name)
            user_arg = cmd_list[pos]
            matched_value = self.service.match_argument_value(name, user_arg)
            if matched_value is None or user_arg != matched_value:
                result.append((arg_name, user_arg))
        return result

    def num_desc_tokens(self, user_tokens):
        """Return number of tokens for given descriptor.

        For commands which don't contain any argument defining limit (and/or
        ending_word), this would simply be len(self.path_tuple). Otherwise
        this will be calculated based on user_tokens, which is command input by
        user so far.

        Args:
            user_tokens: List of command input words as tokens
        TODO: This method is current not being used anywhere and can be removed
        after making sure that it won't be useful for future as well.
        """
        factory = self.service._factory

        # List of descriptors for all items in self.path_tuple
        arg_desc = [factory.arguments.get(resolve_argument_name(arg))
                    for arg in self.path_tuple if is_argument(arg)]

        if not any([d.limit for d in arg_desc]):
            # Command descriptor does not contain any argument with limit
            return len(self.path_tuple)
        else:
            total = 0
            user_tokens_index = 0
            for a in self.path_tuple:
                if not is_argument(a) and user_tokens_index < len(user_tokens):
                    total += 1
                    user_tokens_index += 1
                elif is_argument(a) and user_tokens_index < len(user_tokens):
                    desc = factory.arguments.get(resolve_argument_name(a))

                    if not desc.is_multiarg:
                        total += 1
                        user_tokens_index += 1
                    else:
                        while( (user_tokens_index < len(user_tokens)) and
                                (user_tokens[user_tokens_index] != desc.ending_word)):
                            total += 1
                            user_tokens_index += 1
                        else:
                            total += 1
                            if user_tokens_index == len(user_tokens):
                                user_tokens_index += 1
                else: # user_tokens_index == len(user_tokens)
                    return total # TODO: check logic once again
                    # raise DescriptorException('Unable to determine desc tokens: %s, user_tokens_index: %s' % (a, user_tokens_index))
        if user_tokens_index <= len(user_tokens):
            return total
        else:
            raise DescriptorException('Unable to determine desc tokens')

    def is_command_complete(self, user_tokens):
        command_indexes = self.get_command_indexes(user_tokens)

        if len(command_indexes) == len(self.path_tuple):
            last_arg = self.path_tuple[-1]

            if not is_argument(last_arg):
                return True
            else:
                if last_arg in self.multi_args:
                    arg_name = resolve_argument_name(last_arg)
                    factory = self.service._factory
                    desc = factory.arguments.get(arg_name)
                    if desc.ending_word == user_tokens[-1]:
                        return True
                    else:
                        return False
                else:
                    return True
        else:
            return False


    def ensure_multi_arg_complete(self, arg_name, values):
        factory = self.service._factory
        arg_desc = factory.arguments.get(resolve_argument_name(arg_name))

        if not arg_desc or not arg_desc.is_multiarg: # Complete by default if not multi_arg
            return True

        if arg_desc.ending_word == values[-1]:
            return True
        else:
            return False


    def get_command_literal_argument(self, user_tokens, pos):
        """Return tuple of matching argument/literal, in user provided input.

        get_command_literal_argument(("get", "file"), 0) => ("get", ("get", ))
        get_command_literal_argument(("get", "file"), 1) => ("get", ("file", ))
        TODO: This command require lot of fixes
        """
        command_indexes = self.get_command_indexes(user_tokens)

        c = self.ensure_multi_arg_complete(self.path_tuple[command_indexes[-1][0]], user_tokens)

        if pos >= len(user_tokens) and pos == len(command_indexes) and c: # Position is beyond provided tokens
            if pos >= len(self.path_tuple):
                return ("", ())
            return (self.path_tuple[pos], ())
        if not c: # If multi_arg is not complete, return it
            return (self.path_tuple[command_indexes[-1][0]], ())

        current_pos = 0

        for item in command_indexes:
            total = item[1] # [0,1] => 1, [1, 2] => 2
            start_index = current_pos
            for i in range(total):
                current_pos += 1
                if current_pos > pos:
                    try:
                        return (self.path_tuple[item[0]], user_tokens[start_index:current_pos])
                    except:
                        lg.exception('err', error_code=ec.INTERNAL_ERROR)

        if len(self.path_tuple) <= len(command_indexes):
            raise DescriptorException("Command descriptor '%s' does not match "
                                      "provided input" % str(self.path_tuple))

        return (self.path_tuple[len(command_indexes)], user_tokens[start_index:])


    def get_command_indexes(self, user_tokens):
        """Return list of (index, size) of for each command literal/argument.

        As a command may contain multi_args arguments, this method helps in
        finding starting index of each entered command literal/argument in
        user_tokens and total tokens for each of those items.
        """
        factory = self.service._factory
        results = []
        # size of result is equal to literal/argument to look for in path_tuple
        multi_arg = None

        # Flag to indicate whether current loop over user_tokens should exit
        should_exit = False

        for i, token in enumerate(user_tokens):
            # If OUTPUT_MODIFIER_TOKEN encountered, everything following it can
            # ignored, for the purpose of calculating command indexes.
            if token.startswith(OUTPUT_MODIFIER_TOKEN):
                should_exit = True

            if should_exit:
                break

            # Delaying this check, as loop should break on next iteration
            if token.strip().endswith(OUTPUT_MODIFIER_TOKEN):
                token = token.replace(OUTPUT_MODIFIER_TOKEN, '')
                should_exit = True

            if len(results) >= len(self.path_tuple):
                if not self.additional_last_args and not multi_arg:
                    exc_str = "user_tokens: %s, desc: %s" % (user_tokens, self.additional_last_args)
                    raise DescriptorException("Command overflow: %s" % exc_str)
                elif self.additional_last_args:
                    results[-1][1] = len(user_tokens) - results[-1][0]
                    return results
            if not multi_arg: # else path_item remains same as last
                try:
                    path_item = self.path_tuple[len(results)]
                except IndexError:
                    pass

            if not is_argument(path_item):
                results.append([len(results), 1])
            else:
                path_item_desc = factory.arguments.get(
                                    resolve_argument_name(path_item))
                if not path_item_desc:
                    err_str = ("Unable to find descriptor definition for "
                               "argument: %s" % path_item)
                    lg.error(err_str, error_code=ec.INVALID_ARGUMENT)
                    raise DescriptorException(err_str)
                if not path_item_desc.is_multiarg:
                    results.append([len(results), 1])
                    multi_arg = None
                else:
                    if not multi_arg: # This is first token in multi_arg
                        multi_arg = path_item_desc
                        results.append([len(results), 1])
                    elif token == path_item_desc.ending_word:
                        multi_arg = None
                        results[-1][1] += 1
                    else:
                        results[-1][1] += 1

        return results


    def _resolve_provider(self):
        """Load command provider as specified in command yaml."""
        # Make sure CliCommand yaml has specified either "provider" or
        # "providers" properties, the check for "provider" can be removed
        # when all CliCommand yamls added in Avalanche have been converted
        # to use "providers" instead.
        if ("providers" not in self._raw_schema and
            "provider" not in self._raw_schema):
            msg = "CliCommand '%s' missing providers property" % self.id
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
            raise DescriptorException(msg)
        if "providers" in self._raw_schema:
            providers = self._raw_schema["providers"]
            if not isinstance(providers, type([])):
                msg = ("CliCommand '%s' specified its providers property "
                       "incorrectly: %s" % (self.id, providers))
                lg.error(msg, error_code=ec.INVALID_ARGUMENT)
                assert False, msg
                raise DescriptorException(msg)
            provider_schema = None
            # Find first provider in provider list where node type matches
            provider_any = None
            for provider in providers:
                targets = provider.get("targets", [PROVIDER_TARGET_ANY])
                if self._node_type in targets:
                    provider_schema = provider
                    break
                elif provider_any is None and PROVIDER_TARGET_ANY in targets:
                    provider_any = provider
            # If no providers match node type, allow provider with target any
            if provider_schema is None and provider_any is not None:
                provider_schema = provider_any
            if provider_schema is None:
                return None
        else:
            provider_schema = self._raw_schema["provider"]
        # Make sure provider targets (if specified) matches self._node_type
        provider_targets = provider_schema.get("targets", [PROVIDER_TARGET_ANY])
        if (self._node_type not in provider_targets and
            PROVIDER_TARGET_ANY not in provider_targets):
            return None
        provider_class = self._provider_class[provider_schema["type"]]
        # Assume provider classes are defined in current module
        cls = getattr(sys.modules[__name__], provider_class)
        return cls(self._raw_schema["id"], provider_schema, self)

    def _resolve_renderer(self):
        """Load command output renderer as specified in command yaml."""
        renderer_schema = self._raw_schema["result_text_renderer"]
        renderer_class = self._renderer_class[renderer_schema["type"]]
        # Assume renderer classes are defined in current module
        cls = getattr(sys.modules[__name__], renderer_class)
        return cls(self._raw_schema["id"], renderer_schema)


class InteractiveArgument(object):
    """Class to model interactive arguments.

    Interactive arguments allow CLI commands to prompt the user for
    additional input.  For example a password or "Are you sure (Y/N)?"
    """
    def __init__(self, **fields):
        defaults = {'arg_pattern': None,
                    'password_arg': False,
                    'invalid_arg_msg': None}
        self.__dict__.update(defaults)
        self.__dict__.update(fields)
        self.value = None


class Provider(object):
    """Base class definition for CLI Command providers."""
    def __init__(self, _id, provider_schema, descriptor):
        self._id = _id
        self._provider_schema = provider_schema
        self._descriptor = descriptor
        self._watch_provider = provider_schema.get("watch_provider", False)
        self._watch_process = None

    @property
    def id_(self):
        return self._id

    @property
    def watch_provider(self):
        return self._watch_provider

    @property
    def watch_process(self):
        return self._watch_process

    def invoke(self, cmd_list, context, context_data, modifier_str,
               local_client=True, watch_stdout=None):
        """Sub-classes need to implement this method."""
        assert False, "Derived class %s must define an invoke method" % self

    def show_configuration(self):
        return None


class MockProvider(Provider):
    """A Mock provider that just returns the result property defined in yaml."""
    def __init__(self, _id, provider_schema, descriptor):
        super(MockProvider, self).__init__(_id, provider_schema, descriptor)

    def invoke(self, cmd_list, context, context_data, modifier_str,
               local_client=True, watch_stdout=None):
        return self._provider_schema["result"]


class BaseShellScriptProvider(Provider):
    """A base class for shell script providers."""
    def _invoke(self, cmd, context, context_data, modifier_str,
                local_client=True, watch_stdout=None):
        """Invokes a shell script provider implementation."""
        env = {
            "NSX_CLI_CONTEXT": context,
            "NSX_CLI_CONTEXT_DATA": context_data,
            "NSX_CLI_MODIFIER": modifier_str}
        if os.environ.get("TERM"):
            env["TERM"] = os.environ.get("TERM")
        if self.watch_provider:
            po = None
            try:
                if watch_stdout:
                    po = subprocess.Popen(cmd, stdout=watch_stdout,
                                          stderr=subprocess.STDOUT,
                                          preexec_fn=os.setpgrp, env=env)
                else:
                    po = subprocess.Popen(cmd, env=env)
                self._watch_process = po
                if local_client:
                    while po.poll() is None:
                        time.sleep(.1)
            except KeyboardInterrupt:
                if po:
                    try:
                        po.kill()
                    except OSError:
                        pass
                    # Commands like ping output text on kill, so we wait
                    # a bit to make sure all command output is flushed
                    # before proceeding
                    time.sleep(1)
            finally:
                self._post_watch_invoke(local_client, watch_stdout)
            return None
        po = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, env=env)
        out = po.communicate()[0]
        if sys.version_info >= (3,) and out is not None:
            out = out.decode("utf8")
        return self._prepare_results(out)

    def _post_watch_invoke(self, local_client=True, watch_stdout=None):
        """Derived classes can perform post processing after watch completes."""
        pass

    def _prepare_results(self, out):
        """ Prepare output from shell script provider.

            Shell script providers return output in the form of:
            header\n\nresult, where \n\nresult is optional in cases
            where an error is encountered  Unlike python providers,
            the result portion is plain text.  This method converts
            the result portion to json (if it exists)
        """
        parts = out.split("\n\n", 1)
        if len(parts) != 2:
            return out
        return parts[0] + "\n\n" + simplejson.dumps(parts[1])


class InlineShellScriptProvider(BaseShellScriptProvider):
    """Inline shell providers define what shell commands to run in the yaml."""
    def __init__(self, _id, provider_schema, descriptor):
        super(InlineShellScriptProvider, self).__init__(_id, provider_schema,
                                                        descriptor)
        self._arg_offset = provider_schema.get("script_argument_offset")

    def invoke(self, cmd_list, context, context_data, modifier_str,
               local_client=True, watch_stdout=None):
        script = self._provider_schema["script"]
        cmd = ["/bin/bash", "-c", script]
        if self._arg_offset:
            cmd += cmd_list[self._arg_offset:]
        return self._invoke(
            cmd, context, context_data, modifier_str, local_client,
            watch_stdout)


class LocalShellScriptProvider(BaseShellScriptProvider):
    """Local shell script providers reference a command file to run."""
    def __init__(self, _id, provider_schema, descriptor):
        super(LocalShellScriptProvider, self).__init__(_id, provider_schema,
                                                       descriptor)
        self._arg_offset = provider_schema.get("script_argument_offset")
        self._provider_schema = provider_schema
        self._script = None

    def invoke(self, cmd_list, context, context_data, modifier_str,
               local_client=True, watch_stdout=None):
        if self._script is None:
            path = os.path.join(
                CLI_BASE_DIR, self._provider_schema["script_path"])
            if not os.path.exists(path):
                lg.error(
                    "Local shell script specified by provider %s not found: %s"
                    % (self._id, path), error_code=ec.INVALID_ARGUMENT)
                render_error("MSG_CLI_INTERNAL_ERROR")
                return
            self._script = path
        cmd = ["/bin/bash", self._script]
        if self._arg_offset:
            cmd += cmd_list[self._arg_offset:]
        return self._invoke(
            cmd, context, context_data, modifier_str, local_client,
            watch_stdout)

    def _post_watch_invoke(self, local_client=True, watch_stdout=None):
        """After watch output, add extra newline."""
        if watch_stdout:
            watch_stdout.write("\n")
        else:
            sys.stdout.write("\n")


class LocalPythonProvider(Provider):
    """Local python providers reference a python class (__call__) to invoke."""
    def __init__(self, _id, provider_schema, descriptor):
        super(LocalPythonProvider, self).__init__(_id, provider_schema,
                                                  descriptor)
        self._module = provider_schema["module"]
        self._class_name = provider_schema["class_name"]
        self._provider = None

    def _resolve_provider(self):
        if self._module is None or self._class_name is None:
            render_error("MSG_CLI_INTERNAL_ERROR")
            return False
        try:
            cls = get_class(self._module, self._class_name)
            self._provider = cls(self._descriptor)
            return True
        except (ImportError, AttributeError):
            msg = ("Unable to import python provider '%s': %s.%s" %
                   (self._id, self._module, self._class_name))
            lg.exception(msg, error_code=ec.INTERNAL_ERROR)
            assert False, msg
            # To avoid repeated import attempts, null out module/class_name
            self._module = self._class_name = None
        render_error("MSG_CLI_INTERNAL_ERROR")
        return False

    def invoke(self, cmd_list, context, context_data, modifier_str,
               local_client=True, watch_stdout=None):
        if not self._provider and not self._resolve_provider():
            return None
        return self._provider(cmd_list)

    def show_configuration(self):
        if not self._provider and not self._resolve_provider():
            return None
        return self._provider.show_configuration()

    def get_sort_func(self):
        """Returns provider's class implemented 'sort_key' function"""
        return getattr(self._provider, 'sort_key', None)


class InlineMakoRenderer(object):
    """Inline mako renders define a mako template inline in the yaml."""
    def __init__(self, _id, renderer_schema):
        self._id = _id
        self._renderer_schema = renderer_schema

    def invoke(self, result, modifier_str, sort_key=None, output_prefix=None):
        # Note that due to yaml conversion, the inline template may contain
        # trailing whitespaces, we trim this here though it's possible a
        # provider may want the trailing whitespace, we'll deal with that
        # as necessary in the future
        template = self._renderer_schema["template"].rstrip()
        return render_mako_template_str(template, result, modifier_str,
                                        sort_key, output_prefix)


class LocalMakoFileRenderer(object):
    """Local mako file renders reference a mako template file to use."""
    def __init__(self, _id, renderer_schema):
        self._id = _id
        self._renderer_schema = renderer_schema
        self._template = None

    def invoke(self, result, modifier_str, sort_key=None, output_prefix=None):
        if self._template is None:
            path = os.path.join(
                CLI_BASE_DIR, self._renderer_schema["template_path"])
            if not os.path.exists(path):
                lg.error(
                    "Local template file specified by provider %s not found: %s"
                    % (self._id, path), error_code=ec.INVALID_ARGUMENT)
                render_error("MSG_CLI_INTERNAL_ERROR")
                return
            self._template = path
        with codecs.open(self._template, "r", "utf-8") as fp:
            template_contents = fp.read().strip()
        return render_mako_template_str(template_contents, result,
                                        modifier_str, sort_key, output_prefix)
