# ***************************************************************************
# Copyright 2015-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import errno
import nsx_logging
import os
import glob

from cli.utils.command_utils import invoke_command
from cli.utils.rest_utils import local_node_request
from cli.utils.error_code import ErrorCode as ec
from compat23 import compat23_httplib


lg = nsx_logging.getLogger("cli.utils.file_utils")


_FILESTORE = "/var/vmware/nsx/file-store"
_TEMPSTORE = "/var/vmware/nsx/tmp"
_MOVE_TMP_SCRIPT = "/opt/vmware/nsx-cli/bin/scripts/move_tmpfile_to_filestore"
_DEFAULT_FILE_MODE = 0o644


def get_file_store():
    """Return file-store directory (which is owned by the NAPI user)."""
    return _FILESTORE


def get_tmp_store():
    """Return temp-store directory (which is owned by the admin user)."""
    return _TEMPSTORE


def move_file_to_filestore(src_filename, dst_filename, src_dir=_TEMPSTORE,
                           dst_file_mode=_DEFAULT_FILE_MODE,
                           dst_file_group_name=None):
    """Move specified file in specified directory to the filestore directory.

    Args:
        src_filename: Source file filename.
        dst_filename: Destination file filename.
        src_dir: Source directory where to find file. Value has to be "/tmp"
            or _TEMPSTORE.
        dst_file_mode: (Numeric) Mode of dst_filename once moved, e.g. 0644
        dst_file_group_name: Optional name of group when setting ownership,
            if None, www-data is used.

    Returns:
        (True, None) if file moved successfully and source file is no longer
        in src_dir. (False, err_msg_id) if file has not been moved. The source
        file may or may not exist in src_dir if this method returns False.
        err_msg_id may be None for internal errors or invalid argument errors.
    """
    if src_dir not in [_TEMPSTORE, "/tmp"]:
        lg.error("Invalid source directory specified: %s", src_dir,
                 error_code=ec.INVALID_ARGUMENT)
        return (False, None)
    elif "/" in src_filename:
        lg.error("Invalid source filename specified: %s", src_filename,
                 error_code=ec.INVALID_ARGUMENT)
        return (False, None)
    elif "/" in dst_filename:
        lg.error("Invalid destination filename specified: %s", dst_filename,
                 error_code=ec.INVALID_ARGUMENT)
        return (False, None)
    # Invoke _MOVE_TMP_SCRIPT to move specified file to filestore.
    src_dir_key = "nsxtmp" if src_dir == _TEMPSTORE else "tmp"
    src_path = os.path.abspath(os.path.join(src_dir, src_filename))
    try:
        cmd = [
            "/usr/bin/sudo", _MOVE_TMP_SCRIPT, src_dir_key, src_filename,
            dst_filename, oct(dst_file_mode)]
        if dst_file_group_name is not None:
            cmd.append(dst_file_group_name)
        rc, out, err = invoke_command(cmd)
        if rc == 0:
            return (True, None)
        elif rc == 2:  # File exists in filestore
            return (False, "MSG_FILESTORE_ALREADY_EXISTS\n%s" % dst_filename)
        else:
            lg.error(
                "Error moving tmp file %s to filestore: %s: %s", dst_filename,
                rc, err, error_code=ec.OPERATION_FAILED)
    finally:
        delete_file_if_exists(src_path)
    return (False, None)


def delete_filestore_file(filename):
    """Delete specified file from filestore

    Returns:
        True, if DELETE /node/filestore/<filename> successfully submitted;
        otherwise, False.
    """
    try:
        status = local_node_request(
            "DELETE", "/node/file-store/%s" % filename)[0]
        return status == compat23_httplib.OK
    except:
        lg.exception("Error removing filestore file: %s", filename,
                     error_code=ec.OPERATION_FAILED)
        return False


#------------------------------------------------------------------------
# This method is used by KVM node to delete existing packet capture files
# from filestore
#------------------------------------------------------------------------
def delete_filestore_file_kvm(filename):
    """Delete specified file from filestore

    Returns:
        True, if file is deleted successfully, or no such file exists
        otherwise, False.
    """
    try:
        file_path = os.path.abspath(os.path.join(get_file_store(), filename))
        delete_file_if_exists(file_path)
        return True
    except:
        return False


def delete_file_if_exists(file_path):
    """Delete specified file_path, ignore ENOENT exceptions or
       if file_path is None.

    Args:
        file_path: Full path to the file to delete.
    Raises:
        Un-expected Exceptions.
    """
    try:
        os.unlink(file_path)
    except TypeError:
        if file_path is not None:
            raise
    except Exception as ex:
        if ex.errno != errno.ENOENT:  # ignore EAFP
            raise


def get_newest_file(glob_path):
    """Get the latest file, from among the files matching the glob_path

    Args:
        glob_path: Absolute or relative path containing the pattern to match
    Returns:
        Most recent file from the set of files matching the pattern, or
        empty string if the pattern match returns empty set
    """
    files = glob.glob(glob_path)
    if not files:
        return ''
    else:
        return max(files, key=os.path.getctime)
