# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
from stat import ST_MTIME
from cli.utils.error_code import ErrorCode as ec


lg = nsx_logging.getLogger("cli.utils.logging_utils")


CLEAN_EXTENSIONS = [".log", ".tgz"]

_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "fatal": ("Designates very severe events that may cause application to"
              " abort"),
    "error": ("Designates error events that may allow application to continue"
              " running"),
    "warn": "Designates potentially harmful situations",
    "info": "Designates coarse-grained informational messages",
    "debug": ("Designates fine-grained informational messages for debugging"
              " purposes"),
    "trace": "Designates finer-grained informational events than debug"
}


class LoggingLevelArgumentDescription(object):
    """Value description for LoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _LOGGING_LEVEL_DESC.get(value, "NSX logging level argument")

_EDGE_ROUTING_SERVICE_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "debug": ("Designates fine-grained informational messages for debugging"
                  " purposes"),
    "error": ("Designates error events that may allow application to continue"
               " running"),
    "informational": "Designates coarse-grained informational messages",
    "notification": "Normal but significant conditions",
    "warning": "Designates potentially harmful situations",
}


class RoutingServiceLoggingLevelArgumentDescription(object):
    """Value description for EdgeRoutingServiceLoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_SERVICE_LOGGING_LEVEL_DESC.get(
            value, "NSX Routing Service logging level argument")


_EDGE_ROUTING_PLATFORM_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "fatal": ("Designates very severe events that may cause application to"
              " abort"),
    "error": ("Designates error events that may allow application to continue"
              " running"),
    "warn": "Designates potentially harmful situations",
    "info": "Designates coarse-grained informational messages",
    "debug": ("Designates fine-grained informational messages for debugging"
              " purposes"),
}


class RoutingPlatformLoggingLevelArgumentDescription(object):
    """Value description for RoutingPlatformLoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_PLATFORM_LOGGING_LEVEL_DESC.get(
            value, "NSX Routing Platform logging level argument")


_EDGE_ROUTING_SERVICE_LOGGING_DESTINATION_DESC = {
    "file": "Routing service log",
    "syslog": "Syslog"
}


class RoutingServiceLoggingDestinationArgumentDescription(object):
    """Value description for EdgeRoutingServiceLoggingDestinationArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_SERVICE_LOGGING_DESTINATION_DESC.get(
            value, "NSX Routing Service logging destination argument")


def clear_log_directory(log_file_dir, stale_ts):
    """Delete log files that are older than specified timestamp.

    Args:
        log_file_dir: Directory where log files can be found.
        stale_ts: Delete log files last modified prior to this timestamp.
    """
    try:
        log_files = os.listdir(log_file_dir)
        for log_file in log_files:
            _, ext = os.path.splitext(log_file)
            if ext not in CLEAN_EXTENSIONS:
                continue
            full_path = os.path.join(log_file_dir, log_file)
            st = os.stat(full_path)
            last_modified_ts = st[ST_MTIME]
            if last_modified_ts < stale_ts:
                try:
                    os.unlink(full_path)
                except Exception:
                    lg.exception("Error deleting stale log file %s:", full_path,
                                 error_code=ec.INTERNAL_ERROR)
    except Exception:
        lg.exception("Error listing log directory files:",
                     error_code=ec.INTERNAL_ERROR)


def build_structured_data_map(security=False, audit=False, errorCode=0,
                              suppress_syslog_audit_entry=False, username=None):
    """Build structured data for logging records.
    """
    key = "tags"
    sd_map = {key: {}}
    if security:
        sd_map[key]["security"] = "true"
    if audit:
        sd_map[key]["audit"] = "true"
    if errorCode:
        sd_map[key]["errorCode"] = errorCode
    if suppress_syslog_audit_entry:
        sd_map[key]["suppress_syslog_audit_entry"] = True
    if username is not None:
        sd_map[key]["username"] = username
    return sd_map
