# ***************************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import gevent
import simplejson

from functools import partial
from management_api.napi.root.administration.support_bundles.controller_util\
    import import_ccp_module

DEFAULT_TIMEOUT = 60

endpoint = '/tmp/cli_unix_sockets.sock'
nsx_rpc_client = import_ccp_module("vmware.nsx.rpc")
cli_servers = import_ccp_module("vmware.nsx.cli.cli_servers_pb2")
lg = nsx_logging.getLogger("cli.utils.nsx_rpc_utils")

def get_result(req):
    '''
    Build rpc client stub and return json formatted result.

    :param req: RPC request that contains all information about rpc call.
    :return: json formatted result.
    '''
    client = CliClient().cli_client
    if client is None:
        return "Cannot get CliServer client"
    rsp = client.CallRpc(req, DEFAULT_TIMEOUT)
    client.Destroy()

    if rsp.success:
        return simplejson.loads(rsp.result)
    else:
        raise Exception("Cannot get CliServer response")


def process_result(result):
    """Generates response based on the jmx result
    Args:
        result: Jmx response
    Return:
        String response of the form 'STATUS_CODE\n\nJSON_DATA'
    """
    undefined = "<none>"
    if 'success' not in result:
        lg.error("Unexpected response, 'success' element is not present",
                 error_code=ec.OPERATION_FAILED)
        return "MSG_INVALID_RESPONSE"
    elif result['success'] == False:
        error = result.get('error', {})
        code = error.get('code', undefined)
        if code == 2:
            return "OK\n\n%s" % simplejson.dumps(None)
        message = error.get('message', undefined)
        lg.error("%s" % error, error_code=ec.OPERATION_FAILED)
        return "MSG_REQUEST_FAILED\n%s" \
               % ("(error code: %s, message: %s)" \
                  % (code, normalize(message)))
    elif result['success'] != True:
        lg.error("Unexpected value of 'success' element: %s"
                 % normalize(result['success']), error_code=ec.OPERATION_FAILED)
        return "MSG_INVALID_RESPONSE"
    else:
        return "OK\n\n%s" % simplejson.dumps(result)


def process_result_to_table(result):
    """Generates table response based on the jmx result
    Args:
        result: Jmx JSON response
    Return:
        String response of the form 'STATUS_CODE\n\nTABLE'
    """
    undefined = "<none>"
    if 'success' not in result:
        lg.error("Unexpected response, 'success' element is not present",
                 error_code=ec.OPERATION_FAILED)
        return "MSG_INVALID_RESPONSE"
    elif result['success'] == False:
        error = result.get('error', {})
        code = error.get('code', undefined)
        if code == 2:
            return "OK\n\n%s" % simplejson.dumps(None)
        message = error.get('message', undefined)
        lg.error("%s" % error, error_code=ec.OPERATION_FAILED)
        return "MSG_REQUEST_FAILED\n%s" \
               % ("(error code: %s, message: %s)" \
                  % (code, normalize(message)))
    elif result['success'] != True:
        lg.error("Unexpected value of 'success' element: %s"
                 % normalize(result['success']), error_code=ec.OPERATION_FAILED)
        return "MSG_INVALID_RESPONSE"
    else:
        return "OK\n\n%s" % simplejson.dumps(parse_jsobj_to_table(result))


def parse_jsobj_to_table(jsobj):
    output = ""
    data_list = jsobj.get("data")
    if data_list is None:
        return output
    format_dict = jsobj.get("format")
    header_list = format_dict.get("header")
    widthMap = format_dict.get("widthMap")
    line_format = ""
    for column_name in header_list:
        width = widthMap.get(column_name)
        line_format += "{:<%d}" % width
    line_format += "\n"
    output += line_format.format(*header_list)
    for data in data_list:
        line_data = []
        for field_name in header_list:
            line_data.append(data.get(field_name))
        output += line_format.format(*line_data)
    return output


def lazy_property(function):
    ATTR_NAME = '__' + function.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, ATTR_NAME):
            setattr(self, ATTR_NAME, function(self))
        return getattr(self, ATTR_NAME)
    return _lazy_property

class CliClient:
    @lazy_property
    def cli_client(self):
        client = None
        last_exception = None
        for attempt in range(10):
            try:
                client = nsx_rpc_client.NsxRpcClient(cli_servers.CliService_Stub,
                         "unix://" + endpoint)
                break
            except Exception as e:
                last_exception = e
                lg.info("CliServers NsxRpcClient failed to connect: %s, attempt %s"
                         % (e, attempt + 1))
                gevent.sleep(3)
        else:
            raise last_exception
        return client
