# ***************************************************************************
# Copyright 2014-2018, 2021 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
from datetime import datetime
import errno
import fcntl
import glob
import nsx_logging
import os
import re
import signal
import socket
from subprocess import PIPE
from subprocess import Popen
import stat
import sys
import tarfile
import time

from cli.utils.command_utils import invoke_command
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import CAPTURE_STARTED
from cli.utils.constants import CAPTURE_FAILED
from cli.utils.constants import CAPTURE_INVALID_INTERFACE
from cli.utils.constants import CAPTURE_INTERRUPT_CTRL_C
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.error_utils import get_message
from cli.utils.file_utils import delete_file_if_exists
from cli.utils.file_utils import delete_filestore_file
from cli.utils.file_utils import delete_filestore_file_kvm
from cli.utils.file_utils import get_file_store
from cli.utils.file_utils import get_tmp_store
from cli.utils.file_utils import move_file_to_filestore
from cli.utils.node_utils import get_node_type
from cli.utils.error_code import ErrorCode as ec

g_node_type = get_node_type()
if g_node_type == NODE_TYPE_HOST_ESX:
    from esx_dvs_utils import get_switch_port_from_vif
    from esx_dvs_utils import get_switch_port_from_dvport
    from esx_dvs_utils import get_switch_port_from_port_client_name


lg = nsx_logging.getLogger("cli.utils.pcap_utils")

devnull = open(os.devnull, 'w')

_PCAP_TMP_DIR = get_tmp_store()
if g_node_type == NODE_TYPE_HOST_ESX:
    _PCAP_TMP_DIR = "/tmp"

_PCAP_LOCKS = {}  # key=interface id, val=lockfile file descriptor
_PCAP_POPENS = {}  # key=interface id,
                   # val=(Popen object of tcpdump process, filestore filename)
_LCK_SUFFIX = ".lck" # lock file suffix

_CMD_TCPDUMP = "/usr/sbin/tcpdump"
_CMD_PKTCAP_ESX = "/bin/pktcap-uw"
_CMD_TCPDUMP_ESX = "/bin/tcpdump-uw"
_CMD_TERMTCPDUMP = "/opt/vmware/nsx-cli/bin/scripts/termtcpdump"
_SUDO = "/usr/bin/sudo"

_CAPTURE_POLL_INTERVAL_SECS = 0.1
_CAPTURE_MAX_WAIT_FOR_ERROR_SECS = 0.5
_CAPTURE_DEFAULT_SNAPLEN = 65535
_CAPTURE_DISK_CAPACITY_THRESHOLD = 70

PKTCAP_EXPRESSION_KEYWORDS_MAPS = {
    "ethtype": ["--ethtype"],
    "srcmac": ["--srcmac"],
    "dstmac": ["--dstmac"],
    "mac": ["--mac"],
    "ipproto": ["--proto"],
    "ip": ["--ip"],
    "srcip": ["--srcip"],
    "dstip": ["--dstip"],
    "ipproto6": ["-6", "--proto"],
    "ip6": ["-6", "--ip"],
    "srcip6": ["-6", "--srcip"],
    "dstip6": ["-6", "--dstip"],
    "vni": ["--vni"],
    "vlan": ["--vlan"],
    "port": ["--tcpport"],
    "srcport": ["--srcport"],
    "dstport": ["--dstport"]
}


PKTCAP_EXPRESSION_KEYWORDS_PCAP_MAPS = {
    "ethtype": ["ether", "proto"],
    "srcmac": ["ether", "src"],
    "dstmac": ["ether", "dst"],
    "mac": ["ether", "host"],
    "ipproto": ["ip", "proto"],
    "ip": ["ip", "host"],
    "srcip": ["ip", "src"],
    "dstip": ["ip", "dst"],
    "ipproto6": ["ip6", "proto"],
    "ip6": ["ip6", "host"],
    "srcip6": ["ip6", "src"],
    "dstip6": ["ip6", "dst"],
    "vni": ["geneve", "vni"],
    "vlan": ["vlan"],
    "port": ["tcp", "port"],
    "srcport": ["tcp", "src", "port"],
    "dstport": ["tcp", "dst", "port"]
}

assert list(PKTCAP_EXPRESSION_KEYWORDS_PCAP_MAPS.keys()) == \
        list(PKTCAP_EXPRESSION_KEYWORDS_MAPS.keys())


class ExpressionComposeError(Exception):
    pass


def disk_stat_esx(path):
    sub = Popen("vdf | grep 'tmp' | awk '{print $5}' | awk -F'%' '{print $1}'", shell=True, stdout=PIPE)
    percent_str = sub.stdout.read()
    sub.wait()
    percent = int(percent_str)
    return percent


def disk_stat_nonesx(path):
    disk = os.statvfs(path)
    percent = (disk.f_blocks - disk.f_bfree) * 100 / (disk.f_blocks -disk.f_bfree + disk.f_bavail) + 1
    return percent


def disk_stat(node_type, path, move):
    """
    Check the capacity usage of the directory "path", according to the "node_type"
    """
    percent = 0
    path = os.path.dirname(path)
    if node_type == NODE_TYPE_HOST_ESX:
        percent = disk_stat_esx(path)
    else:
        percent = disk_stat_nonesx(path)
        if not move:
            percent2 = disk_stat_nonesx(get_file_store())
            percent = max(percent, percent2)
    return percent


def _valid_new_expression_syntax(expression):
    """
    Determine if expression is in new syntax.
    We iterates the tokens in expression: if we find a new syntax, then it is
    done. if we aren't able to find one, it must be old syntax.

    New syntax is a key-value pair expression:
    <key> <value> [<key> <value>] [<key> <value>] ...

    For a description of the keys and values:
    ethtype:   ethernet proto, 0x<ETHTYPE>, maximum value 0xFFFF, aka 65535
    srcmac:    source ethernet address, same requirement as ethtype
    dstmac:    destination ethernet address, same requirement as ethtype
    mac:       ethernet address, only accept ee:ee:ee:ee:ee:ee format
    ipproto:   IPv4 proto, 0x<PROTO>, maximum value 0xFF, aka 255
    ip:        IPv4 address
    srcip:     source IPv4 address, same requirement as ip
    dstip:     destination IPv4 address, same requirement as ip
    ipproto6:  IPv6 proto, 0x<PROTO>, maximum value 0xFF, aka 255
    ip6:       IPv6 address
    srcip6:    source IPv6 address, same requirement as ip6
    dstip6:    destination IPv6 address, same requirement as ip6
    vni:       Virtual Network Interface, maximum value 0xffffff, aka 16777215
    vlan:      VLAN tag, maximum value 0xfff, 4095
    port:      TCP protocol port, maximum value 0xffff, aka 65535
    srcport:   TCP protocol source port, same requirement with port
    dstport:   TCP protocol destination port, same requirement with port

    Args:
         expression: expression string
    Return:
         True/False
    """

    tokens = expression.split(" ")
    while len(tokens) > 1:
        if tokens[0] in ["ip", "srcip", "dstip"]:
            try:
                # "ip" <ipv4 address>, must be new syntax
                socket.inet_pton(socket.AF_INET, tokens[1])
                return True
            except socket.error:
                pass
            except ValueError:
                pass
            # other format must be old syntax
            break
        elif tokens[0] in ["ip6", "srcip6", "dstip6"]:
            try:
                # "ip" <ipv6 address>, must be new syntax
                socket.inet_pton(socket.AF_INET6, tokens[1])
                return True
            except socket.error:
                pass
            except ValueError:
                pass
            # other format must be old syntax
            break
        elif tokens[0] in ["mac", "srcmac", "dstmac"]:
            if re.match("^[a-fA-F0-9]{2}(:[a-fA-F0-9]{2}){5}$", tokens[1]):
                return True
            break
        elif tokens[0] in ["ipproto", "ipproto6"]:
            if re.match("^0x[a-fA-F0-9]{1,2}$", tokens[1]):
                # the value must be equal to or less than 0xff
                return True
            break
        elif tokens[0] in ["ethtype"]:
            if re.match("^0x[a-fA-F0-9]{1,4}$", tokens[1]):
                # the value must be equal to or less than 0xffff
                return True
            break
        elif tokens[0] in ["port", "srcport", "dstport"]:
            # "port"    <TCPORT>, must be new syntax
            # "srcport" <SRCPORT>, must be new syntax
            # "dstport" <DSTPORT>, must be new syntax
            if re.match("^[0-9]+$", tokens[1]):
                try:
                    value = int(tokens[1])
                    if value < 0 or value > 0xffff:
                        raise ValueError
                    return True
                except ValueError:
                    pass
            break
        elif tokens[0] in ["vni"]:
            # "vni" <VNI>, must be new syntax
            if re.match("^[0-9]+$", tokens[1]):
                try:
                    value = int(tokens[1])
                    if value < 0 or value > 0xffffff:
                        raise ValueError
                    return True
                except ValueError:
                    pass
            break
        elif tokens[0] in ["vlan"]:
            # "vlan" <vlan>
            if not re.match("^[0-9]+$", tokens[1]):
                break
            try:
                value = int(tokens[1])
                if value < 0 or value > 0xfff:
                    raise ValueError
            except ValueError:
                return False
            # can"t tell if it is new syntax, continue to scan
            tokens = tokens[2:]
        else:
            # doesn"t start with any know types in keywords
            break
    if not tokens:
        return True
    return False


def _convert_expression_to_arguments(expression, maps, seperator = None,
      has_vni_filter_support=True):
    """
    Convert expression to arguments used in command line.
    If the syntax is incorrect, it will raise ExpressionComposeError.

    Args:
         expression: expression string
    Return:
         cmd arguments in an array
    """
    tokens = expression.split(" ")
    first_token = True
    arguments = []
    while len(tokens) >= 2:
        if tokens[0] in maps:
            if not first_token:
                if seperator:
                    arguments.append(seperator)
            else:
                first_token = False
            if tokens[0] == "vni" and not has_vni_filter_support:
                raise ExpressionComposeError(
                      "no vni filter available on this interface."
                      " Please check the log for more information.")
            arguments.extend(maps[tokens[0]])
            arguments.append(tokens[1])
        else:
            raise ExpressionComposeError("bad token '%s', expecting one of %s" % \
                  (tokens[0], list(maps.keys())))
        tokens = tokens[2:]
    if tokens:
        raise ExpressionComposeError("bad token '%s', expecting one of %s" % \
              (tokens[0], list(maps.keys())))
    return arguments


def _is_esx_pnic_get_geneve_software_simulation_cap(pnic_name):
    try:
        import vmkctl
        info = vmkctl.PnicSoftwareSimulationInfo()
        vmkctl.PnicImpl(pnic_name).GetPnicSoftwareSimulationInfo(info)
        if not hasattr(info, "geneveOffload"):
            return False
        return info.geneveOffload
    except:
        return False


def _is_esx_pnic_get_geneve_offload_cap(pnic_name):
    try:
        import vmkctl
        hwcaps = vmkctl.PnicImpl(pnic_name).GetHardwareCapabilities()
        if not hasattr(hwcaps, "GetGeneveOffload"):
            return False
        return hwcaps.GetGeneveOffload()
    except:
        return False


def _is_esx_65():
    from os import uname
    major, minor, _ = uname().release.split('.')
    if int(major) * 10000 + int(minor) * 100 == 60500:
        return True
    return False


def _is_esx_main():
    from os import uname
    major, minor, _ = uname().release.split('.')
    if int(major) * 10000 + int(minor) * 100 > 60700:
        return True
    return False

def get_nsx_agent_style_direction(direction):
    DIRECTION_MAPS = {
        "input": "in",
       "output": "out",
       "dual": "both"
    }
    return DIRECTION_MAPS.get(direction, 'both')

def get_tcpdump_style_direction(direction):
    DIRECTION_MAPS = {
        "input": "in",
       "output": "out",
       "dual": "inout"
    }
    return DIRECTION_MAPS.get(direction, 'inout')


# TODO: determine interfaces via a REST API call to /node/network/interfaces
def _get_local_interfaces():
    """Retrieve available network interfaces.

    Return:
        List of available network interfaces.
    """
    results = []
    # TODO: get list of interfaces from GET /node/network/interfaces
    rc, out, _ = invoke_command(["/bin/netstat", "-i", "-a"])
    if rc == 0:
        lines = out.split("\n")
        if len(lines) > 2:
            lines = lines[2:]
            for line in lines:
                if not line:
                    continue
                parts = line.split()
                results.append(parts[0])
            results.sort()
    return results


# ----------------------------------------------------------------------
# lock, unlock and test interface pcap lock
# ----------------------------------------------------------------------
def lock_interface(interface_id):
    """Grab lock for interface_id.

    Args:
        interface_id: Interface id of lock to grab.
    Return:
        True if lock was successful; otherwise, False.
    """
    lockfile_path = os.path.join(_PCAP_TMP_DIR, interface_id + _LCK_SUFFIX)
    try:
        fd = open(lockfile_path, "w")
        # fcntl.flock raises an IOError exception if the lockfile is
        # already locked by the current process or a different process
        fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        _PCAP_LOCKS[interface_id] = fd
        return True
    except IOError as ex:
        # Note: it's possible that _PCAP_LOCKS[interface_id] exists at this
        # point. For example if within the same CLI session the user tries
        # the perform a capture on the same interface (without stopping the
        # first capture first).
        if ex.errno == errno.EWOULDBLOCK:
            # It's possible that the previous capture already completed on
            # its own, for example if a packet count was specified. Check
            # _PCAP_POPENS to see if this is the case, if so, cleanup and
            # re-check the current lock_interface request
            tcpdump_po, capture_file = _PCAP_POPENS.get(
                interface_id, (None, None))
            if tcpdump_po and tcpdump_po.poll() is not None:
                stop_packet_capture_to_file(interface_id)
                unlock_interface(interface_id)
                return lock_interface(interface_id)
        elif ex.errno not in [errno.EACCES, errno.EAGAIN, errno.EDEADLOCK]:
            lg.exception("Unexpected error while attempting to get pcap lock",
                         error_code=ec.OPERATION_FAILED)
    return False


def unlock_interface(interface_id):
    """Release lock for interface_id.

    Args:
        interface_id: Interface id of lock to release.
    """
    fd = _PCAP_LOCKS.get(interface_id)
    if fd is None:
        return
    del _PCAP_LOCKS[interface_id]
    try:
        fcntl.flock(fd, fcntl.LOCK_UN)
    except Exception:
        pass
    finally:
        try:
            fd.close()
        except Exception:
            pass


def interface_locked(interface_id):
    """Determine if current process has interface locked.

    Args:
        interface_id: Interface id to check whether lock exists.
    """
    return interface_id in _PCAP_LOCKS


# ----------------------------------------------------------------------
# start packet capture to display or file
# *** this function is also used by edge/ops/packet_capture ***
# *** this function is also used by host_kvm/ops/packet_capture ***
# *** this function is also used by host_esx/ops/packet_capture ***
# ----------------------------------------------------------------------
def start_packet_capture(interface_id, capture_to_file=True, expression=None,
                         capture_file=None, packet_count=None,
                         direction=None, dvfilter_name=None,
                         dvfilter_stage=None, snaplen=None, file_count=None):
    """Start a packet capture on specified interface.

    Args:
        interface_id: Interface on which to initiate packet capture.
        capture_to_file: True, if capture to .pcap file; False if display only.
        expression: Filter expression to apply.
        capture_file: Name of destination filestore file, if None, use
            "%s.pcap" % interface_id.
        packet_count: Number of packets to capture before exiting.
        file_count: Number of files of 100M each for rotation (currently only supported on Edge/PCG)
    Return:
        (CAPTURE_STARTED, None), if packet capture started successfully.
        (CAPTURE_INVALID_INTERFACE, None), If an invalid interface_id specified.
        (CAPTURE_FAILED, err_msg), if an error occurred when starting the capture,
                                   err_msg will have tcpdump message for user, or
                                   None if a generatic non-tcpdump error occurred.
    """
    # In case users copy/paste capture_file that contain leading and
    # trailing quotes, strip it off
    if (capture_file and len(capture_file) > 1 and capture_file[0] in ["'", '"'] and
            capture_file[-1] in ["'", '"']):
        capture_file = capture_file[1:-1]

    if snaplen is None:
        snaplen = _CAPTURE_DEFAULT_SNAPLEN

    # Construct the command.  Since tcpdump requires root access to perform
    # the capture, we have to invoke sudo to escalate prvileges.
    g_node_type = get_node_type()

    sigint_handler = None
    sigterm_handler = None
    sigpipe_handler = None

    start_time = None
    err = b""

    if g_node_type is NODE_TYPE_HOST_ESX:
        # ESX always assume UW is running under root
        cmd = [ _CMD_PKTCAP_ESX ]
        if not _is_esx_65():
            # allow vprobe, required by Enhanced Network Stack
            cmd += ["-e"]
        if dvfilter_name:
            cmd += ["--dvfilter", dvfilter_name]
            if dvfilter_stage == "pre":
                cmd += ["--stage", "0"]
            elif dvfilter_stage == "post":
                cmd += ["--stage", "1"]
        elif interface_id == "trace":
            # trace is a special interface_id
            cmd += ["--trace"]
        elif interface_id.startswith("vmk"):
            cmd += ["--vmk", interface_id]
        elif interface_id.startswith("vmnic"):
            cmd += ["--uplink", interface_id, "-K"]
        elif interface_id.startswith("uplink"):
            dvPortId = interface_id
            # uplink alias is actually a dvsport id
            switch_port = get_switch_port_from_dvport(dvPortId)
            if not switch_port:
                return CAPTURE_INVALID_INTERFACE, None
            cmd += ["--switchport", str(switch_port)]
        else:
            # try vif id first
            vifId = interface_id
            switch_port = get_switch_port_from_vif(vifId)
            # try to guess user's input is logical_port_id/dvPortId
            if not switch_port:
                dvPortId = interface_id
                switch_port = get_switch_port_from_dvport(dvPortId)
                # try to guess user's input is port client
            if not switch_port:
                port_client = interface_id
                switch_port = get_switch_port_from_port_client_name(port_client)
            # try to guess user's input is pure switch port
            if not switch_port and interface_id.isdigit():
                switch_port = interface_id
            if not switch_port:
                return CAPTURE_INVALID_INTERFACE, None
            cmd += ["--switchport", str(switch_port)]
        if direction:
            if direction == "input":
                cmd += ["--dir", "0"]
            elif direction == "output":
                cmd += ["--dir", "1"]
            elif direction == "dual":
                cmd += ["--dir", "2"]
    else:
        if interface_id not in _get_local_interfaces():
            return CAPTURE_INVALID_INTERFACE, None
        extra_params = []
        if g_node_type in [
                NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
            start_time = time.time()
            if capture_file is None:
                extra_params = ["-P"]
            # opsagent passes filename as '-' but capture must still go to stdout
            elif file_count:
                # rotate upto 25 files, every 100MB.
                # Maximum disk space used by raw pcap files per caputre will be 2.5GB,
                # the compression should reduce actual disk space further.
                extra_params = ["-W", file_count, "-C", "100"]
        cmd = [_SUDO, _CMD_TCPDUMP, "-i", interface_id, "-n", "-e"] + extra_params
        # tcpdump direction is not needed on span interface, stripping it.
        if not interface_id.startswith("span-"):
            cmd += [ "--direction=%s" % get_tcpdump_style_direction(direction) ]

    pcap_arguments = []

    # Avoid write to temporary file then move.
    # It might cause confusing console output about output name in console
    # described in PR2042106.
    write_to_file_directly = False
    if capture_file == "-" or g_node_type == NODE_TYPE_HOST_ESX:
        write_to_file_directly = True

    if packet_count:
        pcap_arguments += ["-c", str(packet_count)]

    if g_node_type is NODE_TYPE_HOST_ESX:
        cmd.extend(["-P"])

    # Add capture to file arugments if output file specified
    if capture_to_file:
        if write_to_file_directly:
            tmp_pcap_file = capture_file
            if g_node_type == NODE_TYPE_HOST_ESX and capture_file != "-":
                tmp_pcap_file = os.path.join(_PCAP_TMP_DIR, capture_file)
        else:
            tmp_pcap_file = _output_tempfile(interface_id)

        # We need check the capacity of directory to store the cap file
        # Refer to PR 2288939, the short-term solution
        # If the capture is not writen to file, there would be a file movement
        try:
            if capture_file != "-":
                percent = disk_stat(g_node_type, tmp_pcap_file, move=write_to_file_directly)
                if percent > _CAPTURE_DISK_CAPACITY_THRESHOLD:
                    lg.exception("Disk capacity is insufficient, usage is higher than 70%")
                    return (CAPTURE_FAILED, "Disk capacity is insufficient, usage is higher than 70%")
        except OSError as err:
            lg.exception("Failed to checking the capacity, {0}".format(err))
            return (CAPTURE_FAILED, "Failed to checking the capacity, {0}".format(err))

        pcap_arguments.extend(["-w", tmp_pcap_file])
        # unbuffer output to stdout
        if tmp_pcap_file == "-":
            pcap_arguments.append("-U")
        pcap_arguments += ["-s", str(snaplen)]

    # Add optional filter expression if specified
    if expression:
        # In case users copy/paste expressions that contain leading and
        # trailing quotes, strip it off
        if (len(expression) > 1 and expression[0] in ["'", '"'] and
                expression[-1] in ["'", '"']):
            expression = expression[1:-1]

    if expression:
        if g_node_type is NODE_TYPE_HOST_ESX:
            # Host ESX only accepts new syntax natively in upstream versions,
            # otherwise it fallbacks to a software-emulated mode.
            is_new_syntax = _valid_new_expression_syntax(expression)
            has_vni_filter_support = False
            if interface_id.startswith("vmnic"):
                # for pNic, there is vni informationS set.
                # However, that information isn't made publicly available and thus
                # it cannot be extracted via tcpdump-uw.
                lg.info("Attempting to capture pNic... %s" % interface_id)
                has_hw_vni_offload_support = _is_esx_pnic_get_geneve_software_simulation_cap(interface_id)
                has_sw_vni_offload_support = _is_esx_pnic_get_geneve_offload_cap(interface_id)
                lg.info("--- Detected HW CAP set: %d" % has_hw_vni_offload_support)
                lg.info("--- Detected SW CAP set: %d" % has_sw_vni_offload_support)
            else:
                # for vNic, there is not such vni information -- XXX follow vNIC to
                # pNIC binding to look up such information.
                lg.info("Attempting to capture vNic... %s" % interface_id)

            if _is_esx_main():
                lg.info("Use native filters")
                try:
                    arguments = _convert_expression_to_arguments(expression,
                          maps = PKTCAP_EXPRESSION_KEYWORDS_MAPS,
                          has_vni_filter_support=has_vni_filter_support)
                except ExpressionComposeError as e:
                    # Note: we have verified the argument, it is unlikely to happen
                    return (CAPTURE_FAILED, "syntax error: %s" % e)
                cmd += arguments
            else:
                lg.info("Use software-emulated filters")
                pcap_arguments.append("--")
                try:
                    pcap_filters = _convert_expression_to_arguments(expression,
                          maps = PKTCAP_EXPRESSION_KEYWORDS_PCAP_MAPS,
                          seperator = 'and',
                          has_vni_filter_support=has_vni_filter_support)
                except ExpressionComposeError as e:
                    return (CAPTURE_FAILED, "syntax error: %s" % e)
                pcap_arguments += pcap_filters
        else:
            # Fix for #1404492: before adding the user specified
            # expression, add the "--" argument which tells tcpdump to
            # stop accepting options. This ensures users can't craft
            # a malicious tcpdump command.
            pcap_arguments.append("--")
            # Host KVM and Edge can accept new syntax and pcap-filter syntax
            # Try new syntax first, or pcap-filter syntax
            is_new_syntax = _valid_new_expression_syntax(expression)
            # Host KVM and Edge do have such vni information available and a more recent
            # tcpdump.
            # Maybe it will works because all packets are aggregated into single thread
            # at kni inteface with mutiple thread mode enabled.
            has_sw_vni_filter_support = True
            if is_new_syntax:
                try:
                    pcap_filters = _convert_expression_to_arguments(expression,
                          maps = PKTCAP_EXPRESSION_KEYWORDS_PCAP_MAPS,
                          seperator = 'and',
                          has_vni_filter_support = has_sw_vni_filter_support)
                except ExpressionComposeError as e:
                    return (CAPTURE_FAILED, "syntax error: %s" % e)
                pcap_arguments += pcap_filters
            else:
                pcap_arguments.append(expression)

    if g_node_type is not NODE_TYPE_HOST_ESX:
        cmd += pcap_arguments

    lg.info("Invoking: %s" % cmd)
    # This is a display request, output to stdout
    po = None
    try:
        # In ESXi, pipe to tcpdump-uw for better console output
        ps = None
        pipefile = '/tmp/pktcap-%d.pipe' % os.getpid()
        if g_node_type == NODE_TYPE_HOST_ESX:
            # [pktcap-uw] (stdout) ---> [ps] -> [po] ---> (stdin) [tcpdump-wu] -->
            #
            # Avoid using shell pipeline due to potential security issues
            #
            # "Replacing shell pipeline" in Python Official Documentation
            # https://docs.python.org/3/library/subprocess.html#replacing-shell-pipeline
            os.mkfifo(pipefile)
            cmd += [ '-o', pipefile ]
            ps = Popen(cmd, stdout=PIPE, stderr=devnull, bufsize=0)

            cmd = [ _CMD_TCPDUMP_ESX, '-n', '-e', '-r', pipefile ]
            cmd += pcap_arguments
            lg.info("Piping to: %s via %s" % (cmd, pipefile))
            po = Popen(cmd, stdin=ps.stdout, stdout=None, stderr=PIPE, bufsize=0)
            # allow ps is able to receive SIGPIPE signal if po exits
            ps.stdout.close()
        else:
            # python3 assumes bufsize is -1, which is not the same with python2
            po = Popen(cmd, stdout=None, stderr=PIPE, bufsize=0)
        if capture_to_file:
            # If tcpdump detects an error, it will abort fairly quickly, so
            # we wait up to _CAPTURE_MAX_WAIT_FOR_ERROR_SECS seconds to see
            # if tcpdump exits. Note that another reason tcpdump can exit is
            # if packet_count number of packets were captured.
            tcpdump_exited = False
            max_wait_ts = time.time() + _CAPTURE_MAX_WAIT_FOR_ERROR_SECS
            while time.time() < max_wait_ts:
                rc = po.poll()
                if rc is not None:
                    tcpdump_exited = True
                    break
                time.sleep(_CAPTURE_POLL_INTERVAL_SECS)
            if not tcpdump_exited:
                if packet_count:
                    msg = get_message(
                        "MSG_OPS_PACKET_CAPTURE_COUNT_TERMINATE", packet_count)
                else:
                    msg = get_message("MSG_OPS_PACKET_CAPTURE_TERMINATE")
                    if file_count:
                        msg += "Capture will be rotated upto %s files, with maximum size of 100MB per file." % file_count
                sys.stdout.write("%s\n" % msg)

        # In Popen.wait call, parent's process isn't able to do any signal handling.
        # Forward related signals to tcpdump processes and SIGINT, SIGTERM, SIGPIPE will
        # be handled in tcpdump/pktcap-uw processes.
        def _forward_signal_safely(pid, intnum):
            lg.info("Received signal %d", intnum)
            try:
                os.kill(pid, intnum)
            except:
                pass
        sigint_handler = signal.signal(signal.SIGINT, lambda intnum, frame: _forward_signal_safely(po.pid, intnum))
        sigterm_handler = signal.signal(signal.SIGTERM, lambda intnum, frame: _forward_signal_safely(po.pid, intnum))
        sigpipe_handler = signal.signal(signal.SIGPIPE, lambda intnum, frame: _forward_signal_safely(po.pid, intnum))

        err = po.communicate()[1]
    except KeyboardInterrupt:
        if capture_to_file:
            sys.stdout.write("\n")
    except Exception:
        lg.exception("Error invoking: %s" % " ".join(cmd),
                     error_code=ec.INTERNAL_ERROR)
    finally:
        if po:
            try:
                # Ignore KeyboardInterrupt while we perform cleanup and rename
                # temporary capture file, if specified
                signal.signal(signal.SIGINT, signal.SIG_IGN)
                signal.signal(signal.SIGQUIT, signal.SIG_IGN)
                signal.signal(signal.SIGTERM, signal.SIG_IGN)
                if ps:
                    # Forward SIGINT to parent process instead
                    # due to the unstable SIGPIPE deliverying
                    _forward_signal_safely(ps.pid, signal.SIGINT)
                    ps.wait()
                if sys.version_info >= (3,) and err is not None:
                    err = err.decode("utf8")
                if os.path.exists(pipefile):
                    os.unlink(pipefile)
                if po.returncode == 0:
                    # tcpdump completed successfully, output packet summary
                    # (which tcpdump outputs to stderr) to the CLI stdout
                    output_stdout = False
                    for line in err.split("\n"):
                        if (not output_stdout and
                                line.endswith("packets captured")):
                            output_stdout = True
                        if output_stdout:
                            sys.stdout.write("%s\n" % line)
                    if capture_to_file:
                        capture_file = capture_file or "%s.pcap" % interface_id
                        if not write_to_file_directly:
                            _move_capture_file_to_filestore(
                                interface_id, capture_file, start_time)
                    return CAPTURE_STARTED, None
                # An error has occurred. Extract the tcpdump error msg, if it
                # exists, and return this for output to the user. In case of
                # generic (non-tcpdump) error just return None and the NSX CLI
                # will display a generic error msg.
                err_msg = None
                if err.startswith("reading from file"):
                    err_msg = "\n".join(err.splitlines()[1:])
                    if "Interrupted system call" in err_msg and capture_to_file:
                        lg.info(err_msg)
                        out_msg = "pcap file has been generated " \
                                  "successfully at %s" % tmp_pcap_file
                        return CAPTURE_INTERRUPT_CTRL_C, out_msg
                    elif "Interrupted system call" in err_msg and not capture_to_file:
                        lg.info(err_msg)
                        out_msg = None
                        return CAPTURE_INTERRUPT_CTRL_C, out_msg
                if (err.startswith("tcpdump: ") or
                        err.startswith("/usr/sbin/tcpdump: ")):
                    err_msg = err.split("tcpdump: ", 1)[1]
                elif (err.startswith("tcpdump-uw: ") or
                        err.startswith("/bin/tcpdump-uw: ")):
                    err_msg = err.split("tcpdump-uw: ", 1)[1]
                lg.info("pcap error, rc: %s, err: %s, expression: %s",
                        po.returncode, err, expression)
                return CAPTURE_FAILED, err_msg
            finally:
                # restore previous signal handler
                if sigint_handler is not None:
                    signal.signal(signal.SIGINT, sigint_handler)
                if sigterm_handler is not None:
                    signal.signal(signal.SIGTERM, sigterm_handler)
                if sigpipe_handler is not None:
                    signal.signal(signal.SIGPIPE, sigpipe_handler)
    return CAPTURE_FAILED, None


# DEPRECATED: delete after QE updates test cases to use new syntax
# ----------------------------------------------------------------------
# stop packet capture to file
# *** this function is also used by edge/ops/packet_capture ***
# *** this function is also used by host_kvm/ops/packet_capture ***
# ----------------------------------------------------------------------
def stop_packet_capture_to_file(interface_id):
    """Stop an ongoing packet capture to file operation on specified interface.

    Args:
        interface_id: Stop captures on this interface
    Returns:
        True if successful; otherwise, False.
    """
    result = True
    g_node_type = get_node_type()
    tcpdump_po, capture_file = _PCAP_POPENS.pop(interface_id, (None, None))
    # Check if tcpdump started for this interface
    if tcpdump_po is None:
        return False
    # If tcpdump hasn't exited already, kill it
    if g_node_type == NODE_TYPE_HOST_ESX and tcpdump_po.poll() is None:
        # we're root, just kill it
        os.kill(tcpdump_po.pid, signal.SIGTERM)
    elif tcpdump_po.poll() is None:
        # Since tcpdump requires root access to invoke it, we need to escalate
        # to root in order to kill it.
        cmd = [_SUDO, _CMD_TERMTCPDUMP, _CMD_TCPDUMP, interface_id]
        # start_packet_capture always adds the "-s 65535" argument if the
        # capture_to_file option is True (i.e. capture to file instead of to
        # display). Therefore we add "-s 65535" to the optional search term so
        # only capture_to_file sessions are stopped. We should not stop capture
        # to display sessions that were started with:
        # "debug packet display interface <interface-id> [<expression>]"
        other_tcpdump_args = "-s 65535"
        cmd.append(other_tcpdump_args)
        err = ""
        try:
            lg.info("Invoking: %s" % cmd)
            rc, _, err = invoke_command(cmd)
            if rc != 0:
                lg.error("Error stopping tcpdump, status: %s, err: %s", rc, err,
                         error_code=ec.OPERATION_FAILED)
        except Exception:
            result = False
            lg.exception("Error stopping packet capture, err: '%s'", err,
                         error_code=ec.INTERNAL_ERROR)
    # Invoke communicate() on the Popen object that initiated the
    # tcpdump process (which we've just killed using
    # _CMD_TERMTCPDUMP). This needs to be done, otherwise tcpdump
    # shows up as a defunct process in the process table.
    if tcpdump_po:
        try:
            tcpdump_po.communicate()
        except Exception:
            pass  # Ignore errors here, we're just doing cleanup
        _move_capture_file_to_filestore(interface_id, capture_file)
        # If Popen.poll() returns None, then tcpdump is still running
        if tcpdump_po.poll() is None:
            result = False
    return result


# DEPRECATED: delete after QE updates test cases to use new syntax
def stop_all_packet_capture_to_file():
    """Stop all active packet captures to file.

    Return:
        True, if all packet captures to file have been stopped;
        otherwise, False.
    """
    error = False
    for interface_id in _get_local_interfaces():
        if interface_locked(interface_id):
            try:
                result = stop_packet_capture_to_file(interface_id)
                if not result:
                    error = True
            except Exception:
                error = True
            finally:
                unlock_interface(interface_id)
    return not error


# ----------------------------------------------------------------------
# capture file utilities
# ----------------------------------------------------------------------
def _output_tempfile(interface_id):
    """Return full path of temporary .pcap file for specified interface.

    Args:
        interface_id: Interface on which to initiate packet capture.
    Return:
        Full path of temporary .pcap file.
    """

    assert get_node_type() != NODE_TYPE_HOST_ESX

    return os.path.join(_PCAP_TMP_DIR, "%s.pcap" % interface_id)


def _move_capture_file_to_filestore(interface_id, capture_file, start_time=None):
    """Move pcap file from temporary directory to filestore.

    Args:
        interface_id: Interface id of pcap file to move.
        capture_file: Name of destination filestore file.
        start_time: timestamp capture was started
    Return:
        True if move succeeded; otherwise, False.
    """
    tmp_pcap_file = _output_tempfile(interface_id)

    assert get_node_type() != NODE_TYPE_HOST_ESX

    try:
        g_node_type = get_node_type()
        # If specified filename already exists in filestore, delete it.
        if ((g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
            (g_node_type == NODE_TYPE_HOST_SLESKVM) or
            (g_node_type == NODE_TYPE_HOST_RHELKVM) or
            (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
            (g_node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER) or
            (g_node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER) or
            (g_node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER) or
            (g_node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER) or
            (g_node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER) or
            (g_node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER)):
            if not delete_filestore_file_kvm(capture_file):
                return False
        else:
            if not delete_filestore_file(capture_file):
                return False
        # Move the new .pcap file to the filestore
        try:
            src_dir, src_filename = os.path.split(tmp_pcap_file)
            if g_node_type in [
                    NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
                    NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                # On edge, there could be multiple pcap files due to file-count,
                # in that case, create a single tarball.
                match_regex = src_filename + "*"
                match_path = os.path.join(src_dir, match_regex)
                match_files = glob.glob(match_path)
                if len(match_files) > 1 and start_time:
                    hostname = socket.getfqdn()
                    if not hostname:
                        hostname = socket.gethostname()
                    parent_dir = capture_file
                    src_filename = src_filename + ".tgz"
                    capture_file = capture_file + ".tgz"
                    if not delete_filestore_file(capture_file):
                        return False
                    sys.stdout.write("Creating %s, this may take few minutes.\n" % capture_file)
                    tar_filepath = os.path.join(src_dir, src_filename)
                    start_timestamp = datetime.fromtimestamp(start_time).strftime("%Y%m%d_%H%M%S%f")
                    with tarfile.open(tar_filepath, "w:gz") as tar:
                        for f in match_files:
                            _, f_name = os.path.split(f)
                            # the path inside tar is capture_file/hostname/start_timestamp/interface.pcapN
                            tar_f_path = os.path.join(parent_dir, hostname, start_timestamp, f_name)
                            tar.add(f, arcname=tar_f_path)
                    # delete original files
                    for f in match_files:
                        delete_file_if_exists(f)
                    tmp_pcap_file = tar_filepath
                else:
                    # if this is the only file, name may end with pcap, pcap0 or pcap00
                    # get the right file name from matched files to move to filestore.
                    for f in match_files:
                        _, src_filename = os.path.split(f)
            result, _ = move_file_to_filestore(
                src_filename, capture_file, src_dir)
            return result
        except Exception:
            lg.exception("Error moving pcap file '%s' to filestore",
                         tmp_pcap_file, error_code=ec.INTERNAL_ERROR)
    finally:
        # Make sure temporary .pcap file no longer exists
        # even if the move failed
        delete_file_if_exists(tmp_pcap_file)
    return False


# ----------------------------------------------------------------------
# display packet capture file
# ----------------------------------------------------------------------
def display_packet_capture_file(capture_file, options):
    """using tcpdump command to display the given packet capture file

        Args:
        capture_file: full path of packet capture file.
        options: tcpdump command options.
    """
    cmd_list = [_SUDO, _CMD_TCPDUMP, "-r", capture_file]
    if options:
        for index in range(0, len(options)):
            cmd_list.append(options[index])
    invoke_command(cmd_list, display=True)
