# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import platform
import queue
import signal
import time
import traceback


class SignalWatcher:
    def __init__(self):
        self.exit_now = False
        signal.signal(signal.SIGINT, self._set_kill)
        signal.signal(signal.SIGTERM, self._set_kill)
        signal.signal(signal.SIGABRT, self._set_kill)

    def _set_kill(self, signum, frame):
        self.exit_now = True



def main():
    if platform.system() == "VMkernel":
        import spicy
        spicy.set_process_name('nsx-sha')

    import sha.preconfig
    logger = sha.preconfig.get_sub_logger(__name__)

    try:
        logger.info("import forkServer component")
        from sha.forkserver import ForkServer
        logger.info("Starting ForkSever")
        forkserver_monitor, forkserver_pid = ForkServer.start()

        logger.info("importing SHA configuration")
        import sha.config

        # update watchdog pid file, including both SHA main pid and forkserver pid
        if sha.config.SYSTEM != sha.preconfig.PLATFORM_WINDOWS:
            if sha.config.SYSTEM == sha.preconfig.PLATFORM_LINUX:
                if sha.config.node_type_is(sha.config.DIST_UA, sha.config.DIST_GM):
                    pidfile = "/var/run/vmware/nsx-sha/watchdog-nsx-sha.BG.PID"
                else:
                    pidfile = "/var/run/vmware/nsx-netopa/watchdog-nsx-sha.BG.PID"
            elif sha.preconfig.SYSTEM == sha.preconfig.PLATFORM_ESX:
                pidfile = "/var/run/vmware/watchdog-nsx-sha.BG.PID"
            pids = str(os.getpid()) + ' ' + str(forkserver_pid)
            logger.info("Update {}: {}".format(pidfile, pids))
            with open(pidfile, "w") as f:
                f.write(pids)

        import gevent.monkey
        gevent.monkey.patch_all(socket=False, thread=False, subprocess=False)

        logger.info("importing ConfigMonitor")
        from sha.config_monitor import ConfigMonitor

        logger.info("importing DataWhiteboard component")
        from sha.core.whiteboard import DataWhiteboard

        logger.info("importing MetricCollector component")
        from sha.core.collector import MetricCollector
        from sha.core.collector import Collector

        logger.info("starting ChannelHub")
        from sha.core.channel.hub_holder import HubHolder
        HubHolder.start_hubs()

        logger.info("importing LoopNotifier and MonitorNotifier component")
        from sha.core.notifier import LoopNotifierV1, MonitorNotifierV1

        logger.info("importing Plugin component")
        from sha.core.plugin._plugin import Plugin
        from sha.core.plugin import PluginStatusMonitor
        from sha.core.plugin import PluginStatus
        from sha.core.plugin.custom_plugin_notifier_manager import CustomPluginNotifierManager
        from sha.core.plugin.plugin_status_collection import PluginStatusCollection

        logger.info("importing profile resolver")
        from sha.core.profile import Resolver

        logger.info("starting ConfigMonitor")
        config_monitor = ConfigMonitor()
        config_monitor.start()

        nestdb_client = None
        if not (sha.config.is_ua() or sha.config.is_gm()):
            logger.info("importing ShaNestdbClient component")
            from sha.core.profile._nestdb_client import ShaNestdbClient
            logger.info("starting ShaNestdbClient")
            nestdb_client = ShaNestdbClient(
                "system-health-agent"
            )
            nestdb_client.start()

        logger.info("starting DataWhiteboard")
        DataWhiteboard.start()

        logger.info("starting MetricCollector")
        metric_collector = MetricCollector()
        metric_collector.start()

        logger.info("starting notifiers")
        notifiers = []
        monitors = {}
        for plugin_path, plugin_name in sha.config.PRE_INSTALL_SCRIPTS.items():
            try:
                plugin = Plugin.predefined_plugin(plugin_path)
                PluginStatusCollection.add_plugin_status(plugin.id, plugin.name)
                manifest = plugin.manifest
                if manifest.get("MONITOR_TYPE", None) == sha.config.PLUGIN_TYPE_LISTEN:
                    monitor_name = manifest.get("MONITOR", None)
                    if not monitor_name:
                        logger.warning(
                            'Not found monitor name for {}'.format(plugin_path))
                        continue
                    # Use new data whiteboard based framework
                    logger.info("starting monitor {} with data whiteboard based framework"
                                .format(monitor_name))
                    if not monitor_name in monitors:
                        monitor = Collector.get_monitor(monitor_name)
                        monitor.internal_monitor.notify = None
                        monitors[monitor_name] = monitor

                    notifiers.append(MonitorNotifierV1(plugin, manifest.get("metrics", [])))
                    notifiers[-1].start()
                    time.sleep(2)

                else:
                    notifiers.append(LoopNotifierV1(plugin))
                    notifiers[-1].start()
                    time.sleep(2)
            except Exception as e:
                trace = traceback.format_exc()
                logger.warning("Plugin {} loads failed, exception:{}, trace:{}".format(plugin_path, e, trace))

        for monitor in monitors.values():
            try:
                monitor.start()
            except Exception as e:
                trace = traceback.format_exc()
                logger.warning('Monitor {} loads failed, exception:{}, trace:{}'.format(monitor.__class__.name, e, trace))

        plugin_status_monitor = PluginStatusMonitor(HubHolder.mpa_hub())
        plugin_status_monitor.start()
        logger.info("PluginStatusMonitor started")

        watcher = SignalWatcher()
        while not watcher.exit_now and not forkserver_monitor.exit_now():
            time.sleep(1)

        if nestdb_client:
            nestdb_client.stop()
        plugin_status_monitor.stop()
        for notifier in notifiers:
            notifier.stop()
        for monitor in monitors.values():
            monitor.stop()
        CustomPluginNotifierManager.stop_all_notifiers()
        metric_collector.stop()
        DataWhiteboard.stop()
        HubHolder.stop_hubs()
        config_monitor.stop()
        forkserver_monitor.stop()

        logger.info('waiting for thread exiting.')
        for notifier in notifiers:
            notifier.join()
        for monitor in monitors.values():
            monitor.join()

        if nestdb_client:
            nestdb_client.join()
        metric_collector.wait_stop()
        DataWhiteboard.wait_stop()
        HubHolder.wait_hubs_stop()
        config_monitor.join()
        forkserver_monitor.wait_stop()
        logger.info('all threads exited.')

    except Exception as e:
        trace = traceback.format_exc()
        logger.error("Error in main Exception:{}, trace:{}".format(e, trace))
    finally:
        # noinspection PyProtectedMember
        os._exit(0)


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        print(e)
        traceback.print_exc()
    finally:
        # noinspection PyProtectedMember
        os._exit(0)
