# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import traceback
import sha.config

from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache

vdr_stats_map = {
    'RX Unicast Packets on the interface'       :'pktRxUnicast',
    'RX Unicast Bytes on the interface'         :'bytesRxUnicast',
    'TX Unicast Packets on the interface'       :'pktTxUnicast',
    'TX Unicast Bytes on the interface'         :'bytesTxUnicast',
    'TX Policy Unicast Packets on the interface':'pktPolicyTxUnicast',
    'TX Policy Unicast Bytes on the interface'  :'bytesPolicyTxUnicast',
    'RX Broadcast Packets on the interface'     :'pktRxBroadcast',
    'RX Broadcast Bytes on the interface'       :'bytesRxBroadcast',
    'TX Broadcast Packets on the interface'     :'pktTxBroadcast',
    'TX Broadcast Bytes on the interface'       :'bytesTxBroadcast',
    'RX Multicast Packets on the interface'     :'pktRxMulticast',
    'RX Multicast Bytes on the interface'       :'bytesRxMulticast',
    'TX Multicast Packets on the interface'     :'pktTxMulticast',
    'RX Packets System Error on interface'      :'pktSystemError',
    'TX Ref Errors on the interface'            :'pktTxRefError',
    'Packets Deferred Free on the interface'    :'deferredFree',
    'RX Packets Dropped on interface'           :'droppedRx',
    'TX Packets Dynamic ResPool on interface'   :'pktTxDynResPool',

    # LIF IPv4 Net Statistics (approx.):
    'IP & ARP packets RX'                       :'pktRx',
    'IP & ARP packets TX'                       :'pktTx',
    'IP packets Forwarded to Lif'               :'pktForward',
    'IP packets Consumed'                       :'pktConsumed',
    'IP packets Fragmented'                     :'pktFragmented',
    'IP packets Ignored'                        :'pktIgnored',
    'ARP Request RX'                            :'arpReqRx',
    'ARP Request TX'                            :'arpReqTx',
    'ARP Response RX'                           :'arpRspRx',
    'ARP Response TX'                           :'arpRspTx',
    'ARP Request for Proxy RX'                  :'proxyArpSrcNotLif',
    'ARP Request for Proxy My IP RX'            :'proxyArpReq',
    'GARP RX'                                   :'garpRx',
    'GARP TX'                                   :'garpTx',
    'ARP Probes TX'                             :'arpProbeTx',
    'ICMP Echo Req RX'                          :'icmpEchoRx',
    'ICMP Echo Rsp TX'                          :'icmpEchoTx',
    'ICMP Time Exceeded TX'                     :'icmpTtlExpTx',
    'ICMP Net Unreachable TX'                   :'icmpDestNetUnreach',
    'ICMP Host Unreachable TX'                  :'icmpDestHostUnreach',
    'TTL Zero Drops'                            :'dropsTtl',
    'Bad Checksum Drops'                        :'dropsCsum',
    'Unsupported Packets Drops'                 :'dropsUnSupported',
    'Blackholed Packets Drops'                  :'dropsBlackholed',
    'Arp HoldPkts Drops'                        :'dropsHoldPkt',
    'RPF Failure Drops'                         :'dropsRpf',
    'Packet Allocation Failure'                 :'allocFail',
    'Route not found to Dest'                   :'errNoRoute',
    'Neighbor not found'                        :'nbrNotFound',

    #LIF IGMP Statistics (approx.):
    'IGMP Query packets RX'                     :'igmpQueryRx',
    'IGMP Report packets RX'                    :'igmpReportRx',
    'IGMP Leave packets RX'                     :'igmpLeaveRx',
    'IGMP Query packets Forwarded'              :'igmpQueryForward',
    'IGMP Report packets Forwarded'             :'igmpReportForward',
    'IGMP Unsupported RX'                       :'igmpUnsupported',
    'IGMP drops'                                :'igmpDrops',

     # LIF IPv6 Net Statistics (approx.):
    'IPv6 packets RX'                           :'ipv6PktRx',
    'IPv6 packets TX'                           :'ipv6PktTx',
    'IPv6 packets Forwarded to Lif'             :'ipv6PktForward',
    'IPv6 packets Consumed'                     :'ipv6PktConsumed',
    'IPv6 packets Ignored'                      :'ipv6PktIgnored',
    'NDP RS RX'                                 :'ipv6RsRx',
    'NDP RA TX'                                 :'ipv6RaTx',
    'NDP NS RX'                                 :'ipv6NsRx',
    'NDP NS TX'                                 :'ipv6NsTx',
    'NDP NA RX'                                 :'ipv6NaRx',
    'NDP NA TX'                                 :'ipv6NaTx',
    'Unsolicited NA RX'                         :'ipv6UnsolicitedNaRx',
    'Unsolicited NA TX'                         :'ipv6UnsolicitedNaTx',
    'Unicast NS Probes TX'                      :'ipv6NeighProbeTx',
    'ICMPv6 Echo Req RX'                        :'icmpv6EchoRx',
    'ICMPv6 Echo Rsp TX'                        :'icmpv6EchoTx',
    'ICMPv6 Time Exceeded TX'                   :'icmpv6TtlExpTx',
    'ICMPv6 Net Unreachable TX'                 :'icmpv6DestNetUnreach',
    'ICMPv6 Host Unreachable TX'                :'icmpv6DestHostUnreach',
    'IPv6 TTL Zero Drops'                       :'ipv6DropsTtl',
    'IPv6 Bad Checksum Drops'                   :'ipv6DropsCsum',
    'IPv6 Blackholed Packets Drops'             :'ipv6DropsBlackholed',
    'NDP NS HoldPkts Drops'                     :'ipv6DropsHoldPkt',
    'IPv6 RPF Failure Drops'                    :'ipv6DropsRpf',
    'IPv6 Packet Allocation Failure'            :'ipv6AllocFail',
    'IPv6 Route not found to Dest'              :'ipv6ErrNoRoute',
}

class VdrStats(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)

    def get_lif_stats(self, vdr_uuid, lif_uuid):
        # nsxcli seems very slow, need 5~6 seconds per run
        #     nsxcli -c 'get logical-router {} interface {} stats |json'

        cmd_get_lif_stats = ['net-vdr', '-L', '--stats', '-n', lif_uuid, vdr_uuid]
        try:
            output = run_command(cmd_get_lif_stats, shell=False) or ''
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get lif stats {}:{}".format(e, trace))
            return None

        lif_stats = {}
        for line in output.splitlines():
            items = line.split(":")
            if len(items) < 2:
                continue
            key = items[0].strip()
            if key in vdr_stats_map.keys():
                lif_stats[vdr_stats_map[key]] = items[1].strip()
        return lif_stats

    def get_lif_policy_ref(self, vdr_uuid, lif_uuid):
        # net-vdr --policy -l -o lifPtRef -n <lif UUID> <router UUID>
        #
        # DR <uuid> LIF <lif_uuid> Policy Table Reference Information :
        #
        # Policy Table ID: 0
        # Next hop mac: 00:50:56:b9:af:00    <== optional"
        # Policy type: proximity routing
        cmd_get_lif_policy_ref = ['net-vdr', '--policy', '-l', '-o', 'lifPtRef', '-n', lif_uuid, vdr_uuid]
        try:
            output = run_command(cmd_get_lif_policy_ref, shell=False) or ''
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get lif policy ref{}:{}".format(e, trace))
            return None

        lines = output.splitlines()
        next_hop = None
        table_id = None
        policy_type = None
        for line in lines:
            if line.startswith("Policy type:"):
                policy_type = line.split(':')[1].lstrip().rstrip()
            elif line.startswith("Next hop mac:"):
                next_hop = line.split(':', 1)[1].lstrip().rstrip().replace(':', '-')
            elif line.startswith("Policy table ID:"):
                table_id = int(line.split(':')[1])

        if table_id is None or policy_type is None:
            return None
        ref_key = policy_type + '.' + vdr_uuid + '.' + lif_uuid
        if next_hop:
            ref_key += ('.' + next_hop)
        return {ref_key: table_id}

    def get_vdr_stats(self, is_global_policy):
        cmd_get_all_vdr = ['net-vdr', '-I', '-l']

        output = run_command(cmd_get_all_vdr, shell=False) or ''
        vdr_list = set([x.split(':')[1].strip() for x in output.splitlines() if 'DR UUID:' in x])
        self._logger.debug("get vdr list [{}]".format(vdr_list))
        vdr_stats_list = []
        vdr_policy_ref_list = []
        for vdr in vdr_list:
            cmd_get_vdr_state = ['net-vdr', '-I', '-l', vdr]
            try:
                output = run_command(cmd_get_vdr_state, shell=False) or ''
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.info("Failed to get vdr state {}:{}".format(e, trace))
                continue
            vdrState = next(x.split(':')[1].strip() for x in output.splitlines() if 'State:' in x)
            self._logger.debug("get vdrState {}: {}".format(vdr, vdrState))
            vdrStates = vdrState.split(',')
            for state in vdrStates:
                if state == "Enabled Locally Managed":
                    continue

            cmd_get_lif = ['net-vdr', '-L', '-l', vdr]
            try:
                output = run_command(cmd_get_lif, shell=False) or ''
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.info("Failed to get lif {}:{}".format(e, trace))
                continue
            lif_list = set([x.split(':')[1].strip() for x in output.splitlines() if 'UUID:' in x])
            self._logger.debug("get lif list of vdr {}: {}".format(vdr, lif_list))

            lif_stats = {}
            vdr_policy_ref = {}
            for lif in lif_list:
                stats = self.get_lif_stats(vdr, lif)
                if stats:
                   lif_stats[lif] = stats
                # Get policy reference info for the per-table policy.
                if not is_global_policy:
                    policy_ref = self.get_lif_policy_ref(vdr, lif)
                    if policy_ref:
                        vdr_policy_ref.update(policy_ref)
            if lif_stats != {}:
                vdr_stats_list.append({vdr: lif_stats})
            if vdr_policy_ref != {}:
                vdr_policy_ref_list.append(vdr_policy_ref)
        self._logger.debug("vdr stats {}".format(vdr_stats_list))
        self._logger.debug("vdr policy ref {}".format(vdr_policy_ref_list))
        return (vdr_stats_list if vdr_stats_list != [] else None), (vdr_policy_ref_list if vdr_policy_ref_list else None)

    def get_policy_per_table(self, table_id=None):
        # Get policy entry in one table, eg:
        # net-vdr --policy -l -B policyTableId
        #
        # Policy Table
        # Destination      GenMask          Flags    Ref Action   HitCount
        # -----------      -------          -----    --- ------   --------
        # 10.1.1.0         255.255.255.0             1   allow    0
        # Policy count = 1
        #
        # if no table_id specified, get policy entries of the legacy global table
        cmd_get_policy_table = ['net-vdr', '--policy', '-l'] + \
                               (['-B', table_id] if table_id is not None else [])
        try:
            output = run_command(cmd_get_policy_table, shell=False) or ''
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get policy {}:{}".format(e, trace))
            return None
        lines = output.splitlines()

        # Get policycount from the last line of output: "Policy count = 1"
        policyCount = int(lines[-1].split('=')[1])
        if policyCount < 1:
            return None

        vdr_policy_table = {}
        # Skip the first lines, including blank line and the titles
        for i in range(len(lines)):
            items = lines[i].split()
            if (len(items) == 6) and (items[0] == "-----------"):
               start_line = i + 1
               break

        for i in range(start_line, start_line + policyCount):
            items = lines[i].split()
            if len(items) == 5: # No Flags
                vdr_policy_table["{}.{}.{}".format(items[0].replace('.', '-'), items[1].replace('.', '-'), items[3])] = int(items[4])
            elif len(items) == 6: # Flags
                vdr_policy_table["{}.{}.{}".format(items[0].replace('.', '-'), items[1].replace('.', '-'), items[4])] = int(items[5])
            else:
                continue
        self._logger.debug("vdr policy {}".format(vdr_policy_table))
        return vdr_policy_table if vdr_policy_table else None

    def get_policy_tables(self):
        # Get legacy global policy entry firstly
        policy_table = self.get_policy_per_table()
        if policy_table:
            vdr_policies = {"mon_global":policy_table}
            return True, None, vdr_policies

        # List all policy tables:
        # net-vdr --policyTable -l
        #
        # Policy Tables
        # Id  Name                                     Ref
        # --- ------------------------------           ---
        # 1   table1                                   0
        # 2   table2                                   0
        cmd_get_policy_tables = ['net-vdr', '--policyTable', '-l']
        try:
            output = run_command(cmd_get_policy_tables, shell=False) or ''
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get policyTable {}:{}".format(e, trace))
            return False, None, None

        table_list = list([y.split() for y in (x.strip() for x in output.splitlines()[4:]) if y])
        self._logger.debug("vdr policy table list {}".format(table_list))

        vdr_policies = {}
        vdr_tables = {}
        for table in table_list:
            vdr_tables[table[1]] = table[0]
            policy_table = self.get_policy_per_table(table[0])
            if policy_table:
                vdr_policies[table[0]] = policy_table

        self._logger.debug("vdr policy tables {}".format(vdr_tables))
        self._logger.debug("vdr policies {}".format(vdr_policies))
        return False, (vdr_tables if vdr_tables else None), (vdr_policies if vdr_policies else None)

    @property
    def wavefront_source(self):
        # Get hostname as source to export metrics to wavefront
        return sha.config.HOST_NAME.lower()

    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        try:
            is_global_policy, vdr_tables, vdr_policies = self.get_policy_tables()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get vdr policy tables {}:{}".format(e, trace))
            is_global_policy = False
            vdr_tables = None
            vdr_policies = None

        try:
            vdr_stats_list, vdr_policy_ref = self.get_vdr_stats(is_global_policy)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.info("Failed to get vdr stats {}:{}".format(e, trace))
            vdr_stats_list = None
            vdr_policy_ref = None


        return vdr_stats_list, vdr_tables, vdr_policies, vdr_policy_ref, self.wavefront_source
