# Copyright (C) 2020 - 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import traceback
import sha.config
import queue
import time
from threading import Thread
from sha.contrib.monitor.processStatus.internal.stats import Stats

class Client:
    """
    Manage remote clients
    """
    _logger = sha.config.get_logger(__name__)
    _children = {}
    _queue = None
    _stop = True
    _timer_thread = None

    @classmethod
    def set_queue(cls, queue):
        cls._stop = False
        cls._queue = queue
        cls._start_timer()

    @classmethod
    def add_client(cls, client, ns):
        cls._logger.info("PSM:To-be-add client is {}".format(ns))
        if cls._children.get(ns) is not None:
            cls._children[ns].update(client)
        else:
            cls._children[ns] = ClientChildWrapper(ns, client, cls._queue)
            cls._children[ns].start()

    @classmethod
    def del_client(cls, ns):
        cls._logger.info("PSM:To-be-delete client is {}".format(ns))
        if cls._children.get(ns) is not None:
            cls._children[ns].stop()
            del cls._children[ns]
        else:
            cls._logger.warning("PSM:To-be-delete client {} not found".format(ns))

    @classmethod
    def stop(cls):
        cls._stop = True
        for ns, client in cls._children.items():
            client.stop()
        cls._children = {}
        cls._timer_thread.join()

    @classmethod
    def _check_liveness(cls):
        while not cls._stop:
            for ns, client in cls._children.items():
                client.check_liveness()
            time.sleep(1)

    @classmethod
    def _start_timer(cls):
        cls._timer_thread = Thread(target=cls._check_liveness, daemon=True)
        cls._timer_thread.start()

class ClientChildWrapper(object):
    """
    Wrapper of remote client
    """
    def __init__(self, ns, client, queue):
        self._logger = sha.config.get_logger(__name__)
        self._queue = queue
        self._ns = ns
        self._client = client
        self._stats = Stats(self._ns)

    def start(self):
        self._client.update(self._ns, self._stats, self._queue)
        notif = self._stats.start()
        if notif:
            self._logger.debug("PSM:notif array is {}".format(notif))
            self._queue.put(notif)

    def update(self, client):
        self._client = client
        self._client.update(self._ns, self._stats, self._queue)

    def check_liveness(self):
        notif = self._stats.check_liveness()
        if notif:
            self._logger.debug("PSM:notif array is {}".format(notif))
            self._queue.put(notif)

    def stop(self):
        self._client.stop()
        self._stats.stop()


class ClientChild(object):
    """
    Shadow of the remote client
    """
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._ns = None
        self._stats = None

    def update(self, ns, stats, notif_queue):
        self._notif_queue = notif_queue
        self._ns = ns
        self._stats = stats

    def refresh_data(self, stats):
        key = "nsx.process.{}.status.{}".format(self._ns, stats['key'])
        data = {}
        data[key] = {'val': stats['val'], 'timestamp': stats['ts']}
        self._logger.debug("PSM:data array is {}".format(data))
        self._notif_queue.put(data)

    def update_heartbeat_ts(self, ts):
        self._stats.update_heartbeat(ts)

    def stop(self):
        self._stats = None
