# Copyright (C) 2020 - 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import traceback
import sha.config
import time
import uuid
from threading import Lock

HEARTBEAT_EXPIRATION_THRESHOLD = 10

def id_mapper(ns):
    id = 0
    for letter in ns:
        id = id + ord(letter)
    return id

class Stats(object):
    """
    Keep the client's heartbeat and state
    """
    def __init__(self, ns):
        self._logger = sha.config.get_logger(__name__)
        self._ts = None
        self._ns = ns
        self._state = {}
        self._lock = Lock()
        self._key = "nsx.process.{}.liveness".format(self._ns)
        self._eid = uuid.UUID(int = id_mapper(self._ns))

    def start(self):
        notif = {"old_state": None, "new_state": "UP", "eid": self._eid}
        notif_dict = {}
        self._state[self._key] = "UP"
        notif_dict[self._key] = notif
        return notif_dict

    def update_heartbeat(self, ts):
        with self._lock:
            self._ts = ts
            self._logger.debug("PSM:{} heartbeat timestamp is {}".format(self._ns, self._ts))

    def update_state(self, state):
        notif_state = self._state.get(self._key)
        if (state == "DOWN") and (notif_state == "UP"):
            notif = {"old_state": self._state[self._key], "new_state": "DOWN", "eid": self._eid}
            self._state[self._key] = "DOWN"
        elif (state == "UP") and (notif_state == "DOWN"):
            notif = {"old_state": self._state[self._key], "new_state": "UP", "eid": self._eid}
            self._state[self._key] = "UP"
        else:
            notif = None

        notif_dict = {}
        if notif:
            notif_dict[self._key] = notif

        return notif_dict

    def check_liveness(self):
        with self._lock:
            interval = (int(time.monotonic()) - self._ts) if self._ts else None

        state = self._state.get(self._key)
        if interval and interval > HEARTBEAT_EXPIRATION_THRESHOLD:
            sha.config._logger.info("PSM:{} ts is {} interval is {}".format(self._ns, self._ts, interval))
            if state == "UP":
                notif = {"old_state": self._state[self._key], "new_state": "DOWN", "eid": self._eid}
                self._state[self._key] = "DOWN"
            else:
                notif = None
        else:
            if state == "DOWN":
                notif = {"old_state": state, "new_state": "UP", "eid": self._eid}
                self._state[self._key] = "UP"
            else:
                notif = None

        notif_dict = {}
        if notif:
            notif_dict[self._key] = notif

        return notif_dict

    def stop(self):
        self._state = {}
