# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import ctypes
import threading
import time
import traceback

import sha.config
from sha.core.channel.provider.base import BaseProvider
from vmware.nsx.sha import sha_pb2


class MpaProvider(BaseProvider):
    """The MPA Provider"""

    def __init__(self):
        self._name = "MpaProvider"
        self._logger = sha.config.get_logger(__name__)
        self._logger.info("initiating channel provider <{}>".format(self.__class__.name))

        self._logger.info("platform is {}".format(sha.config.SYSTEM))
        if sha.config.SYSTEM == sha.config.PLATFORM_WINDOWS:
            self.sha_client = ctypes.CDLL('sha_client.dll')
        else:
            self.sha_client = ctypes.CDLL('libshaclientshared.so.0')

        self._mutex = threading.Lock()
        self.sha_client_thread = threading.Thread(target=self.sha_client.createNSXShaClient, daemon=True,
                                                  name="nsx-sha-create-nsx-sha-client")
        self.sha_client_thread.start()
        # wait for sha client thread finish creation
        time.sleep(5)

    def sync(self, msg):
        with self._mutex:
            payload = msg.payload
            payload_len = len(payload)
            payload_type = msg.payload_type
            topic = msg.topic
            try:
                self._logger.info("before sendNSXShaMessage, msg:{}, payload_len:{}".format(msg, payload_len))
                self.sha_client.sendNSXShaMessage(payload, payload_len, payload_type, topic)
                self._logger.info("sendNSXShaMessage successfully.")
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("{} failed in sendNSXShaMessage via SHA client, msg:{}, Exception:{}, trace:{}"
                                     .format(self.name, msg, e, trace))

    def close(self):
        self._logger.info("{} is destroying SHA client.".format(self.name))
        with self._mutex:
            try:
                self.sha_client.destroyNSXShaClient()
                del self.sha_client
                self._logger.info("{} finished destroying SHA client.".format(self.name))
            except Exception:
                trace = traceback.format_exc()
                self._logger.warning("{} failed in destroy SHA client, trace:{}".format(self.name, trace))

    @property
    def name(self):
        return self._name


class ReportMsg:
    def __init__(self, topic, payload, payload_type):
        self._topic = topic
        self._payload = payload
        self._payload_type = payload_type

    def __repr__(self):
        return 'ReportMsg: topic:{}, payload:{}, payload_type:{}'\
            .format(self.topic, sha_pb2.ShaReportMsg().FromString(self.payload), self.payload_type)

    @property
    def topic(self):
        return self._topic

    @property
    def payload(self):
        return self._payload

    @property
    def payload_type(self):
        return self._payload_type