# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import copy
import threading
import traceback

import sha.config
from sha.core.plugin.plugin_status import PluginStatus
from sha.core.profile._resolver import Resolver
from sha.constants import ERROR, INSTALL_FAIL


class PluginStatusCollection:
    _logger = sha.config.get_logger(__name__)

    plugin_status_list = {}
    lock = threading.RLock()

    @classmethod
    def add_plugin_status(cls, plugin_id, plugin_name):
        with cls.lock:
            exist_status = cls.plugin_status_list.get(plugin_id, None)
            if exist_status is not None:
                exist_status.name = plugin_name
                cls._logger.warn("Plugin status is already in collection, no need to add again. exist_status:{}"
                                 .format(exist_status))
                return

            cls.plugin_status_list[plugin_id] = PluginStatus(plugin_id, plugin_name)

    @classmethod
    def get_plugin_status(cls, plugin_id):
        return cls.plugin_status_list.get(plugin_id, None)

    @classmethod
    def remove_plugin_status(cls, plugin_id):
        with cls.lock:
            plugin_status = cls.plugin_status_list.pop(plugin_id, None)
            if plugin_status is not None:
                cls._logger.info("Removed plugin status:{}".format(plugin_status))

    @classmethod
    def get_all_plugin_status(cls):
        # refresh profile info, in case some plugin's profile is out of date.
        # e.g. UplinkStatusMonitor (not check by interval) or other plugins whose check interval is large.
        cls._refresh_profile_info()
        return copy.deepcopy(cls.plugin_status_list)

    @classmethod
    def update_plugin_status(cls, plugin_id, status, detail=None, profile=None):
        with cls.lock:
            plugin_status = cls.get_plugin_status(plugin_id)
            if plugin_status:
                plugin_status.status = status
                if detail is not None:
                    plugin_status.detail = detail
                if profile is not None:
                    plugin_status.profile = copy.deepcopy(profile)

    @classmethod
    def update_plugin_status_profile(cls, plugin_id, profile=None):
        with cls.lock:
            plugin_status = cls.get_plugin_status(plugin_id)
            if plugin_status:
                plugin_status.profile = copy.deepcopy(profile)

    @classmethod
    def _refresh_profile_info(cls):
        for status in cls.plugin_status_list.values():
            if status.status == INSTALL_FAIL:
                # For those install_fail plugin, no need to refresh its profile
                continue

            plugin_id = status.plugin_id
            try:
                profile = Resolver.get(plugin_id)
            except Exception as e:
                cls.update_plugin_status(plugin_id, ERROR, "Failed in loading profile." + str(e), {})
                trace = traceback.format_exc()
                cls._logger.warn("Refresh_profile_info - Failed in load profile for plugin id {}, trace:{}"
                                 .format(plugin_id, trace))
                continue
            # TODO update logic here, perform update if any diff exist or error.
            cls.update_plugin_status_profile(plugin_id, profile)
